import numbers
import random
import cv2
import numpy as np
import PIL

import torch
import torchvision

from . import functional as F


class Compose(object):
    """Composes several transforms

    Args:
    transforms (list of ``Transform`` objects): list of transforms
    to compose
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, clip):
        for t in self.transforms:
            clip = t(clip)
        return clip


class RandomHorizontalFlip(object):
    """Horizontally flip the list of given images randomly
    with a probability 0.5
    """

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray

        Returns:
        PIL.Image or numpy.ndarray: Randomly flipped clip
        """
        if random.random() < 0.5:
            if isinstance(clip[0], np.ndarray):
                return [np.fliplr(img) for img in clip]
            elif isinstance(clip[0], PIL.Image.Image):
                return [
                    img.transpose(PIL.Image.FLIP_LEFT_RIGHT) for img in clip
                ]
            else:
                raise TypeError('Expected numpy.ndarray or PIL.Image' +
                                ' but got list of {0}'.format(type(clip[0])))
        return clip


class RandomResize(object):
    """Resizes a list of (H x W x C) numpy.ndarray to the final size

    The larger the original image is, the more times it takes to
    interpolate

    Args:
    interpolation (str): Can be one of 'nearest', 'bilinear'
    defaults to nearest
    size (tuple): (widht, height)
    """

    def __init__(self, ratio=(3. / 4., 4. / 3.), interpolation='nearest'):
        self.ratio = ratio
        self.interpolation = interpolation

    def __call__(self, clip):
        scaling_factor = random.uniform(self.ratio[0], self.ratio[1])

        if isinstance(clip[0], np.ndarray):
            im_h, im_w, im_c = clip[0].shape
        elif isinstance(clip[0], PIL.Image.Image):
            im_w, im_h = clip[0].size

        new_w = int(im_w * scaling_factor)
        new_h = int(im_h * scaling_factor)
        new_size = (new_w, new_h)
        resized = F.resize_clip(
            clip, new_size, interpolation=self.interpolation)
        return resized


class Resize(object):
    """Resizes a list of (H x W x C) numpy.ndarray to the final size

    The larger the original image is, the more times it takes to
    interpolate

    Args:
    interpolation (str): Can be one of 'nearest', 'bilinear'
    defaults to nearest
    size (tuple): (widht, height)
    """

    def __init__(self, size, interpolation='nearest'):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, clip):
        resized = F.resize_clip(
            clip, self.size, interpolation=self.interpolation)
        return resized


class RandomCrop(object):
    """Extract random crop at the same location for a list of images

    Args:
    size (sequence or int): Desired output size for the
    crop in format (h, w)
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            size = (size, size)

        self.size = size

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray

        Returns:
        PIL.Image or numpy.ndarray: Cropped list of images
        """
        h, w = self.size
        if isinstance(clip[0], np.ndarray):
            im_h, im_w, im_c = clip[0].shape
        elif isinstance(clip[0], PIL.Image.Image):
            im_w, im_h = clip[0].size
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image' +
                            'but got list of {0}'.format(type(clip[0])))
        if w > im_w or h > im_h:
            error_msg = (
                'Initial image size should be larger then '
                'cropped size but got cropped sizes : ({w}, {h}) while '
                'initial image is ({im_w}, {im_h})'.format(
                    im_w=im_w, im_h=im_h, w=w, h=h))
            raise ValueError(error_msg)

        x1 = random.randint(0, im_w - w)
        y1 = random.randint(0, im_h - h)
        cropped = F.crop_clip(clip, y1, x1, h, w)

        return cropped


class ThreeCrop(object):
    """Extract random crop at the same location for a list of images

    Args:
    size (sequence or int): Desired output size for the
    crop in format (h, w)
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            size = (size, size)

        self.size = size

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray

        Returns:
        PIL.Image or numpy.ndarray: Cropped list of images
        """
        h, w = self.size
        if isinstance(clip[0], np.ndarray):
            im_h, im_w, im_c = clip[0].shape
        elif isinstance(clip[0], PIL.Image.Image):
            im_w, im_h = clip[0].size
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image' +
                            'but got list of {0}'.format(type(clip[0])))
        if w != im_w and h != im_h:
            clip = F.resize_clip(clip, self.size, interpolation="bilinear")
            im_h, im_w, im_c = clip[0].shape

        step = np.max((np.max((im_w, im_h)) - self.size[0]) // 2, 0)
        cropped = []
        for i in range(3):
            if (im_h > self.size[0]):
                x1 = 0
                y1 = i * step
                cropped.extend(F.crop_clip(clip, y1, x1, h, w))
            else:
                x1 = i * step
                y1 = 0
                cropped.extend(F.crop_clip(clip, y1, x1, h, w))
        return cropped


class RandomRotation(object):
    """Rotate entire clip randomly by a random angle within
    given bounds

    Args:
    degrees (sequence or int): Range of degrees to select from
    If degrees is a number instead of sequence like (min, max),
    the range of degrees, will be (-degrees, +degrees).

    """

    def __init__(self, degrees):
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError('If degrees is a single number,'
                                 'must be positive')
            degrees = (-degrees, degrees)
        else:
            if len(degrees) != 2:
                raise ValueError('If degrees is a sequence,'
                                 'it must be of len 2.')

        self.degrees = degrees

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray

        Returns:
        PIL.Image or numpy.ndarray: Cropped list of images
        """
        import skimage
        angle = random.uniform(self.degrees[0], self.degrees[1])
        if isinstance(clip[0], np.ndarray):
            rotated = [skimage.transform.rotate(img, angle) for img in clip]
        elif isinstance(clip[0], PIL.Image.Image):
            rotated = [img.rotate(angle) for img in clip]
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image' +
                            'but got list of {0}'.format(type(clip[0])))

        return rotated


class CenterCrop(object):
    """Extract center crop at the same location for a list of images

    Args:
    size (sequence or int): Desired output size for the
    crop in format (h, w)
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            size = (size, size)

        self.size = size

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray

        Returns:
        PIL.Image or numpy.ndarray: Cropped list of images
        """
        h, w = self.size
        if isinstance(clip[0], np.ndarray):
            im_h, im_w, im_c = clip[0].shape
        elif isinstance(clip[0], PIL.Image.Image):
            im_w, im_h = clip[0].size
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image' +
                            'but got list of {0}'.format(type(clip[0])))
        if w > im_w or h > im_h:
            error_msg = (
                'Initial image size should be larger then '
                'cropped size but got cropped sizes : ({w}, {h}) while '
                'initial image is ({im_w}, {im_h})'.format(
                    im_w=im_w, im_h=im_h, w=w, h=h))
            raise ValueError(error_msg)

        x1 = int(round((im_w - w) / 2.))
        y1 = int(round((im_h - h) / 2.))
        cropped = F.crop_clip(clip, y1, x1, h, w)

        return cropped


class ColorJitter(object):
    """Randomly change the brightness, contrast and saturation and hue of the clip

    Args:
    brightness (float): How much to jitter brightness. brightness_factor
    is chosen uniformly from [max(0, 1 - brightness), 1 + brightness].
    contrast (float): How much to jitter contrast. contrast_factor
    is chosen uniformly from [max(0, 1 - contrast), 1 + contrast].
    saturation (float): How much to jitter saturation. saturation_factor
    is chosen uniformly from [max(0, 1 - saturation), 1 + saturation].
    hue(float): How much to jitter hue. hue_factor is chosen uniformly from
    [-hue, hue]. Should be >=0 and <= 0.5.
    """

    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0):
        self.brightness = brightness
        self.contrast = contrast
        self.saturation = saturation
        self.hue = hue

    def get_params(self, brightness, contrast, saturation, hue):
        if brightness > 0:
            brightness_factor = random.uniform(
                max(0, 1 - brightness), 1 + brightness)
        else:
            brightness_factor = None

        if contrast > 0:
            contrast_factor = random.uniform(
                max(0, 1 - contrast), 1 + contrast)
        else:
            contrast_factor = None

        if saturation > 0:
            saturation_factor = random.uniform(
                max(0, 1 - saturation), 1 + saturation)
        else:
            saturation_factor = None

        if hue > 0:
            hue_factor = random.uniform(-hue, hue)
        else:
            hue_factor = None
        return brightness_factor, contrast_factor, saturation_factor, hue_factor

    def __call__(self, clip):
        """
        Args:
        clip (list): list of PIL.Image

        Returns:
        list PIL.Image : list of transformed PIL.Image
        """
        if isinstance(clip[0], np.ndarray):
            raise TypeError(
                'Color jitter not yet implemented for numpy arrays')
        elif isinstance(clip[0], PIL.Image.Image):
            brightness, contrast, saturation, hue = self.get_params(
                self.brightness, self.contrast, self.saturation, self.hue)

            # Create img transform function sequence
            img_transforms = []
            if brightness is not None:
                img_transforms.append(lambda img: torchvision.transforms.functional.adjust_brightness(img, brightness))
            if saturation is not None:
                img_transforms.append(lambda img: torchvision.transforms.functional.adjust_saturation(img, saturation))
            if hue is not None:
                img_transforms.append(lambda img: torchvision.transforms.functional.adjust_hue(img, hue))
            if contrast is not None:
                img_transforms.append(lambda img: torchvision.transforms.functional.adjust_contrast(img, contrast))
            random.shuffle(img_transforms)

            # Apply to all images
            jittered_clip = []
            for img in clip:
                for func in img_transforms:
                    jittered_img = func(img)
                jittered_clip.append(jittered_img)

        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image' +
                            'but got list of {0}'.format(type(clip[0])))
        return jittered_clip


class Normalize(object):
    """Normalize a clip with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``

    .. note::
        This transform acts out of place, i.e., it does not mutates the input tensor.

    Args:
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, clip):
        """
        Args:
            clip (Tensor): Tensor clip of size (T, C, H, W) to be normalized.

        Returns:
            Tensor: Normalized Tensor clip.
        """
        return F.normalize(clip, self.mean, self.std)

    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1})'.format(self.mean, self.std)



"""RCNN Model."""
from __future__ import absolute_import

import warnings

import mxnet as mx
from mxnet import gluon
from mxnet.gluon import nn

from ...nn.coder import NormalizedBoxCenterDecoder, MultiPerClassDecoder
from ...nn.feature import FPNFeatureExpander


class RCNN(gluon.HybridBlock):
    """RCNN network.

    Parameters
    ----------
    features : gluon.HybridBlock
        Base feature extractor before feature pooling layer.
    top_features : gluon.HybridBlock
        Tail feature extractor after feature pooling layer.
    classes : iterable of str
        Names of categories, its length is ``num_class``.
    box_features : gluon.HybridBlock
        feature head for transforming roi output for box prediction.
    short : int
        Input image short side size.
    max_size : int
        Maximum size of input image long side.
    train_patterns : str
        Matching pattern for trainable parameters.
    nms_thresh : float
        Non-maximum suppression threshold. You can specify < 0 or > 1 to disable NMS.
    nms_topk : int
        Apply NMS to top k detection results, use -1 to disable so that every Detection
         result is used in NMS.
    roi_mode : str
        ROI pooling mode. Currently support 'pool' and 'align'.
    roi_size : tuple of int, length 2
        (height, width) of the ROI region.
    strides : int/tuple of ints
        Stride(s) of network features. Tuple for FPN.
    clip: float
        Clip bounding box target to this value.

    Attributes
    ----------
    classes : iterable of str
        Names of categories, its length is ``num_class``.
    num_class : int
        Number of positive categories.
    short : int
        Input image short side size.
    max_size : int
        Maximum size of input image long side.
    train_patterns : str
        Matching pattern for trainable parameters.
    nms_thresh : float
        Non-maximum suppression threshold. You can specify < 0 or > 1 to disable NMS.
    nms_topk : int
        Apply NMS to top k detection results, use -1 to disable so that every Detection
         result is used in NMS.
    """

    def __init__(self, features, top_features, classes, box_features, short, max_size,
                 train_patterns, nms_thresh, nms_topk, post_nms, roi_mode, roi_size, strides, clip,
                 force_nms=False, **kwargs):
        super(RCNN, self).__init__(**kwargs)
        self.classes = classes
        self.num_class = len(classes)
        self.short = short
        self.max_size = max_size
        self.train_patterns = train_patterns
        self.nms_thresh = nms_thresh
        self.nms_topk = nms_topk
        self.post_nms = post_nms
        self.force_nms = force_nms

        assert self.num_class > 0, "Invalid number of class : {}".format(self.num_class)
        assert roi_mode.lower() in ['align', 'pool'], "Invalid roi_mode: {}".format(roi_mode)
        self._roi_mode = roi_mode.lower()
        assert len(roi_size) == 2, "Require (h, w) as roi_size, given {}".format(roi_size)
        self._roi_size = roi_size
        self._strides = strides

        with self.name_scope():
            self.features = features
            self.top_features = top_features
            self.box_features = box_features
            self.class_predictor = nn.Dense(
                self.num_class + 1, weight_initializer=mx.init.Normal(0.01))
            self.box_predictor = nn.Dense(
                self.num_class * 4, weight_initializer=mx.init.Normal(0.001))
            self.cls_decoder = MultiPerClassDecoder(num_class=self.num_class + 1)
            self.box_decoder = NormalizedBoxCenterDecoder(clip=clip, convert_anchor=True)

    def collect_train_params(self, select=None):
        """Collect trainable params.

        This function serves as a help utility function to return only
        trainable parameters if predefined by experienced developer/researcher.
        For example, if cross-device BatchNorm is not enabled, we will definitely
        want to fix BatchNorm statistics to avoid scaling problem because RCNN training
        batch size is usually very small.

        Parameters
        ----------
        select : select : str
            Regular expressions for parameter match pattern

        Returns
        -------
        The selected :py:class:`mxnet.gluon.ParameterDict`

        """
        if select is None:
            return self.collect_params(self.train_patterns)
        return self.collect_params(select)

    def set_nms(self, nms_thresh=0.3, nms_topk=400, force_nms=False, post_nms=100):
        """Set NMS parameters to the network.

        .. Note::
            If you are using hybrid mode, make sure you re-hybridize after calling
            ``set_nms``.

        Parameters
        ----------
        nms_thresh : float, default is 0.3.
            Non-maximum suppression threshold. You can specify < 0 or > 1 to disable NMS.
        nms_topk : int, default is 400
            Apply NMS to top k detection results, use -1 to disable so that every Detection
             result is used in NMS.
        force_nms : bool, default is False
            Appy NMS to all categories, this is to avoid overlapping detection results
            from different categories.

        Returns
        -------
        None

        """
        self._clear_cached_op()
        self.nms_thresh = nms_thresh
        self.nms_topk = nms_topk
        self.force_nms = force_nms
        self.post_nms = post_nms

    def reset_class(self, classes, reuse_weights=None):
        """Reset class categories and class predictors.

        Parameters
        ----------
        classes : iterable of str
            The new categories. ['apple', 'orange'] for example.
        reuse_weights : dict
            A {new_integer : old_integer} or mapping dict or {new_name : old_name} mapping dict,
            or a list of [name0, name1,...] if class names don't change.
            This allows the new predictor to reuse the
            previously trained weights specified.

        """
        self._clear_cached_op()
        if reuse_weights:
            assert hasattr(self, 'classes'), "require old classes to reuse weights"
        old_classes = getattr(self, 'classes', [])
        self.classes = classes
        self.num_class = len(classes)
        if isinstance(reuse_weights, (dict, list)):
            if isinstance(reuse_weights, dict):
                # trying to replace str with indices
                for k, v in reuse_weights.items():
                    if isinstance(v, str):
                        try:
                            v = old_classes.index(v)  # raise ValueError if not found
                        except ValueError:
                            raise ValueError(
                                "{} not found in old class names {}".format(v, old_classes))
                        reuse_weights[k] = v
                    if isinstance(k, str):
                        try:
                            new_idx = self.classes.index(k)  # raise ValueError if not found
                        except ValueError:
                            raise ValueError(
                                "{} not found in new class names {}".format(k, self.classes))
                        reuse_weights.pop(k)
                        reuse_weights[new_idx] = v
            else:
                new_map = {}
                for x in reuse_weights:
                    try:
                        new_idx = self.classes.index(x)
                        old_idx = old_classes.index(x)
                        new_map[new_idx] = old_idx
                    except ValueError:
                        warnings.warn("{} not found in old: {} or new class names: {}".format(
                            x, old_classes, self.classes))
                reuse_weights = new_map

        with self.name_scope():
            old_class_pred = self.class_predictor
            old_box_pred = self.box_predictor
            ctx = list(old_class_pred.params.values())[0].list_ctx()
            # to avoid deferred init, number of in_channels must be defined
            in_units = list(old_class_pred.params.values())[0].shape[1]
            self.class_predictor = nn.Dense(
                self.num_class + 1, weight_initializer=mx.init.Normal(0.01),
                prefix=self.class_predictor.prefix, in_units=in_units)
            self.box_predictor = nn.Dense(
                self.num_class * 4, weight_initializer=mx.init.Normal(0.001),
                prefix=self.box_predictor.prefix, in_units=in_units)
            self.cls_decoder = MultiPerClassDecoder(num_class=self.num_class + 1)
            # initialize
            self.class_predictor.initialize(ctx=ctx)
            self.box_predictor.initialize(ctx=ctx)
            if reuse_weights:
                assert isinstance(reuse_weights, dict)
                # class predictors
                srcs = (old_class_pred, old_box_pred)
                dsts = (self.class_predictor, self.box_predictor)
                offsets = (1, 0)  # class predictor has bg, box don't
                lens = (1, 4)  # class predictor length=1, box length=4
                for src, dst, offset, l in zip(srcs, dsts, offsets, lens):
                    for old_params, new_params in zip(src.params.values(),
                                                      dst.params.values()):
                        # slice and copy weights
                        old_data = old_params.data()
                        new_data = new_params.data()

                        for k, v in reuse_weights.items():
                            if k >= len(self.classes) or v >= len(old_classes):
                                warnings.warn("reuse mapping {}/{} -> {}/{} out of range".format(
                                    k, self.classes, v, old_classes))
                                continue
                            new_data[(k + offset) * l:(k + offset + 1) * l] = \
                                old_data[(v + offset) * l:(v + offset + 1) * l]
                        # reuse background weights as well
                        if offset > 0:
                            new_data[0:l] = old_data[0:l]
                        # set data to new conv layers
                        new_params.set_data(new_data)

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, x, width, height):
        """Not implemented yet."""
        raise NotImplementedError


def custom_rcnn_fpn(pretrained_base=True, base_network_name='resnet18_v1b', norm_layer=nn.BatchNorm,
                    norm_kwargs=None, sym_norm_layer=None, sym_norm_kwargs=None,
                    num_fpn_filters=256, num_box_head_conv=4, num_box_head_conv_filters=256,
                    num_box_head_dense_filters=1024):
    r"""Generate custom RCNN model with resnet base network w/FPN.

    Parameters
    ----------
    pretrained_base : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    base_network_name : str, default 'resnet18_v1b'
        base network for mask RCNN. Currently support: 'resnet18_v1b', 'resnet50_v1b',
        and 'resnet101_v1d'
    norm_layer : nn.HybridBlock, default nn.BatchNorm
        Gluon normalization layer to use. Default is frozen batch normalization layer.
    norm_kwargs : dict
        Keyword arguments for gluon normalization layer
    sym_norm_layer : nn.SymbolBlock, default `None`
        Symbol normalization layer to use in FPN. This is due to FPN being implemented using
        SymbolBlock. Default is `None`, meaning no normalization layer will be used in FPN.
    sym_norm_kwargs : dict
        Keyword arguments for symbol normalization layer used in FPN.
    num_fpn_filters : int, default 256
        Number of filters for FPN output layers.
    num_box_head_conv : int, default 4
        Number of convolution layers to use in box head if batch normalization is not frozen.
    num_box_head_conv_filters : int, default 256
        Number of filters for convolution layers in box head.
        Only applicable if batch normalization is not frozen.
    num_box_head_dense_filters : int, default 1024
        Number of hidden units for the last fully connected layer in box head.

    Returns
    -------
    SymbolBlock or HybridBlock
        Base feature extractor eg. resnet w/ FPN.
    None or HybridBlock
        R-CNN feature before each task heads.
    HybridBlock
        Box feature extractor
    """
    use_global_stats = norm_layer is nn.BatchNorm
    if base_network_name == 'resnet18_v1b':
        from ...model_zoo.resnetv1b import resnet18_v1b
        base_network = resnet18_v1b(pretrained=pretrained_base, dilated=False,
                                    use_global_stats=use_global_stats, norm_layer=norm_layer,
                                    norm_kwargs=norm_kwargs)
        fpn_inputs_names = ['layers1_relu3_fwd', 'layers2_relu3_fwd', 'layers3_relu3_fwd',
                            'layers4_relu3_fwd']
    elif base_network_name == 'resnet50_v1b':
        from ...model_zoo.resnetv1b import resnet50_v1b
        base_network = resnet50_v1b(pretrained=pretrained_base, dilated=False,
                                    use_global_stats=use_global_stats, norm_layer=norm_layer,
                                    norm_kwargs=norm_kwargs)
        fpn_inputs_names = ['layers1_relu8_fwd', 'layers2_relu11_fwd', 'layers3_relu17_fwd',
                            'layers4_relu8_fwd']
    elif base_network_name == 'resnet101_v1d':
        from ...model_zoo.resnetv1b import resnet101_v1d
        base_network = resnet101_v1d(pretrained=pretrained_base, dilated=False,
                                     use_global_stats=use_global_stats, norm_layer=norm_layer,
                                     norm_kwargs=norm_kwargs)
        fpn_inputs_names = ['layers1_relu8_fwd', 'layers2_relu11_fwd', 'layers3_relu68_fwd',
                            'layers4_relu8_fwd']
    elif base_network_name == 'resnest50':
        from ...model_zoo.resnest import resnest50
        base_network = resnest50(pretrained=pretrained_base, dilated=False,
                                 use_global_stats=use_global_stats, norm_layer=norm_layer,
                                 norm_kwargs=norm_kwargs)
        fpn_inputs_names = ['layers1_relu11_fwd', 'layers2_relu15_fwd', 'layers3_relu23_fwd',
                            'layers4_relu11_fwd']
    elif base_network_name == 'resnest101':
        from ...model_zoo.resnest import resnest101
        base_network = resnest101(pretrained=pretrained_base, dilated=False,
                                  use_global_stats=use_global_stats, norm_layer=norm_layer,
                                  norm_kwargs=norm_kwargs)
        fpn_inputs_names = ['layers1_relu11_fwd', 'layers2_relu15_fwd', 'layers3_relu91_fwd',
                            'layers4_relu11_fwd']
    else:
        raise NotImplementedError('Unsupported network', base_network_name)
    features = FPNFeatureExpander(
        network=base_network, outputs=fpn_inputs_names,
        num_filters=[num_fpn_filters] * len(fpn_inputs_names), use_1x1=True,
        use_upsample=True, use_elewadd=True, use_p6=True, no_bias=not use_global_stats,
        pretrained=pretrained_base, norm_layer=sym_norm_layer, norm_kwargs=sym_norm_kwargs)
    top_features = None
    box_features = nn.HybridSequential()
    box_features.add(nn.AvgPool2D(pool_size=(3, 3), strides=2, padding=1))  # reduce to 7x7
    if use_global_stats:
        box_features.add(
            nn.Dense(num_box_head_dense_filters, weight_initializer=mx.init.Normal(0.01)),
            nn.Activation('relu'))
    else:
        for _ in range(num_box_head_conv):
            box_features.add(nn.Conv2D(num_box_head_conv_filters, 3, padding=1, use_bias=False),
                             norm_layer(**norm_kwargs),
                             nn.Activation('relu'))
    box_features.add(
        nn.Dense(num_box_head_dense_filters, weight_initializer=mx.init.Normal(0.01)),
        nn.Activation('relu'))
    return features, top_features, box_features



"""The functions for monodepth2 model and loss
Code partially borrowed from
https://github.com/nianticlabs/monodepth2/blob/master/layers.py.
"""
# pylint: disable=unused-argument, missing-function-docstring
# Copyright Niantic 2019. Patent Pending. All rights reserved.
#
# This software is licensed under the terms of the Monodepth2 licence
# which allows for non-commercial use only, the full terms of which are made
# available in the LICENSE file.

from __future__ import absolute_import, division, print_function

import numpy as np

import mxnet as mx
from mxnet.gluon import nn


def disp_to_depth(disp, min_depth, max_depth):
    """Convert network's sigmoid output into depth prediction
    The formula for this conversion is given in the 'additional considerations'
    section of the paper.
    """
    min_disp = 1 / max_depth
    max_disp = 1 / min_depth
    scaled_disp = min_disp + (max_disp - min_disp) * disp
    depth = 1 / scaled_disp
    return scaled_disp, depth


def transformation_from_parameters(axisangle, translation, invert=False):
    """Convert the network's (axisangle, translation) output into a 4x4 matrix
    """
    R = rot_from_axisangle(axisangle)
    t = translation.copy()

    if invert:
        R = R.transpose(axes=(0, 2, 1))
        t = t * -1

    T = get_translation_matrix(t)

    if invert:
        M = mx.nd.batch_dot(R, T)
    else:
        M = mx.nd.batch_dot(T, R)

    return M


def get_translation_matrix(translation_vector):
    """Convert a translation vector into a 4x4 transformation matrix
    """
    t = translation_vector.reshape((-1, 3, 1))

    elem1 = mx.nd.ones(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem2 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem3 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem4 = t[:, 0, :]

    elem5 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem6 = mx.nd.ones(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem7 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem8 = t[:, 1, :]

    elem9 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem10 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem11 = mx.nd.ones(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem12 = t[:, 2, :]

    elem13 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem14 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem15 = mx.nd.zeros(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))
    elem16 = mx.nd.ones(translation_vector.shape[0]).as_in_context(
        context=translation_vector.context).reshape((translation_vector.shape[0], 1))

    elem = [elem1, elem2, elem3, elem4, elem5, elem6, elem7, elem8,
            elem9, elem10, elem11, elem12, elem13, elem14, elem15, elem16]
    T = mx.nd.concat(*elem, dim=1).reshape((translation_vector.shape[0], 4, 4))

    return T


def rot_from_axisangle(vec):
    """Convert an axisangle rotation into a 4x4 transformation matrix
    (adapted from https://github.com/Wallacoloo/printipi)
    Input 'vec' has to be Bx1x3
    """
    angle = mx.nd.norm(vec, 2, 2, keepdims=True)
    axis = vec / (angle + 1e-7)

    ca = mx.nd.cos(angle)
    sa = mx.nd.sin(angle)
    C = 1 - ca

    x = axis[..., 0].expand_dims(axis=1)
    y = axis[..., 1].expand_dims(axis=1)
    z = axis[..., 2].expand_dims(axis=1)

    xs = x * sa
    ys = y * sa
    zs = z * sa
    xC = x * C
    yC = y * C
    zC = z * C
    xyC = x * yC
    yzC = y * zC
    zxC = z * xC

    elem1 = mx.nd.squeeze(x * xC + ca).reshape((vec.shape[0], 1))
    elem2 = mx.nd.squeeze(xyC - zs).reshape((vec.shape[0], 1))
    elem3 = mx.nd.squeeze(zxC + ys).reshape((vec.shape[0], 1))
    elem4 = mx.nd.zeros(vec.shape[0]).as_in_context(
        context=vec.context).reshape((vec.shape[0], 1))
    elem5 = mx.nd.squeeze(xyC + zs).reshape((vec.shape[0], 1))
    elem6 = mx.nd.squeeze(y * yC + ca).reshape((vec.shape[0], 1))
    elem7 = mx.nd.squeeze(yzC - xs).reshape((vec.shape[0], 1))
    elem8 = mx.nd.zeros(vec.shape[0]).as_in_context(
        context=vec.context).reshape((vec.shape[0], 1))
    elem9 = mx.nd.squeeze(zxC - ys).reshape((vec.shape[0], 1))
    elem10 = mx.nd.squeeze(yzC + xs).reshape((vec.shape[0], 1))
    elem11 = mx.nd.squeeze(z * zC + ca).reshape((vec.shape[0], 1))
    elem12 = mx.nd.zeros(vec.shape[0]).as_in_context(
        context=vec.context).reshape((vec.shape[0], 1))
    elem13 = mx.nd.zeros(vec.shape[0]).as_in_context(
        context=vec.context).reshape((vec.shape[0], 1))
    elem14 = mx.nd.zeros(vec.shape[0]).as_in_context(
        context=vec.context).reshape((vec.shape[0], 1))
    elem15 = mx.nd.zeros(vec.shape[0]).as_in_context(
        context=vec.context).reshape((vec.shape[0], 1))
    elem16 = mx.nd.ones(vec.shape[0]).as_in_context(
        context=vec.context).reshape((vec.shape[0], 1))

    elem = [elem1, elem2, elem3, elem4, elem5, elem6, elem7, elem8,
            elem9, elem10, elem11, elem12, elem13, elem14, elem15, elem16]
    rot = mx.nd.concat(*elem, dim=1).reshape((vec.shape[0], 4, 4))

    return rot


class ConvBlock(nn.HybridBlock):
    """Layer to perform a convolution followed by ELU
    """

    def __init__(self, in_channels, out_channels):
        super(ConvBlock, self).__init__()

        with self.name_scope():
            self.conv = Conv3x3(in_channels, out_channels)
            self.nonlin = nn.ELU()

    def hybrid_forward(self, F, x):
        out = self.conv(x)
        out = self.nonlin(out)
        return out

    def predict(self, x):
        out = self.conv.predict(x)
        out = self.nonlin(out)
        return out


class Conv3x3(nn.HybridBlock):
    """Layer to pad and convolve input
    """

    def __init__(self, in_channels, out_channels, use_refl=True):
        super(Conv3x3, self).__init__()
        self.use_refl = use_refl

        with self.name_scope():
            self.pad = nn.ReflectionPad2D(1)
            self.conv = nn.Conv2D(in_channels=int(in_channels),
                                  channels=int(out_channels), kernel_size=3)

    def hybrid_forward(self, F, x):
        if self.use_refl:
            out = self.pad(x)
        else:
            out = F.pad(x, mode='constant', constant_value=0,
                        pad_width=(0, 0, 0, 0, 1, 1, 1, 1))

        out = self.conv(out)
        return out

    def predict(self, x):
        if self.use_refl:
            out = self.pad(x)
        else:
            out = mx.nd.pad(x, mode='constant', constant_value=0,
                            pad_width=(0, 0, 0, 0, 1, 1, 1, 1))

        out = self.conv(out)
        return out


class BackprojectDepth(nn.HybridBlock):
    """Layer to transform a depth image into a point cloud
    """

    def __init__(self, batch_size, height, width, ctx=mx.cpu()):
        super(BackprojectDepth, self).__init__()

        self.batch_size = batch_size
        self.height = height
        self.width = width

        self.ctx = ctx

        meshgrid = np.meshgrid(range(self.width), range(self.height), indexing='xy')
        id_coords = np.stack(meshgrid, axis=0).astype(np.float32)
        id_coords = mx.nd.array(id_coords).as_in_context(self.ctx)

        pix_coords = mx.nd.expand_dims(mx.nd.stack(*[id_coords[0].reshape(-1),
                                                     id_coords[1].reshape(-1)], axis=0),
                                       axis=0)
        pix_coords = pix_coords.repeat(repeats=batch_size, axis=0)
        pix_coords = pix_coords.as_in_context(self.ctx)

        with self.name_scope():
            self.id_coords = self.params.get('id_coords', shape=id_coords.shape,
                                             init=mx.init.Zero(), grad_req='null')
            self.id_coords.initialize(ctx=self.ctx)
            self.id_coords.set_data(mx.nd.array(id_coords))

            self.ones = self.params.get('ones',
                                        shape=(self.batch_size, 1, self.height * self.width),
                                        init=mx.init.One(), grad_req='null')
            self.ones.initialize(ctx=self.ctx)

            self.pix_coords = self.params.get('pix_coords',
                                              shape=(self.batch_size, 3, self.height * self.width),
                                              init=mx.init.Zero(), grad_req='null')
            self.pix_coords.initialize(ctx=self.ctx)
            self.pix_coords.set_data(mx.nd.concat(pix_coords, self.ones.data(), dim=1))

    def hybrid_forward(self, F, depth, inv_K, **kwargs):
        cam_points = F.batch_dot(inv_K[:, :3, :3], self.pix_coords.data())
        cam_points = depth.reshape(self.batch_size, 1, -1) * cam_points
        cam_points = F.concat(cam_points, self.ones.data(), dim=1)

        return cam_points

    def predict(self, depth, inv_K, **kwargs):
        cam_points = mx.nd.batch_dot(inv_K[:, :3, :3], self.pix_coords.data())
        cam_points = depth.reshape(self.batch_size, 1, -1) * cam_points
        cam_points = mx.nd.concat(cam_points, self.ones.data(), dim=1)

        return cam_points


class Project3D(nn.HybridBlock):
    """Layer which projects 3D points into a camera with intrinsics K and at position T
    """

    def __init__(self, batch_size, height, width, eps=1e-7):
        super(Project3D, self).__init__()

        self.batch_size = batch_size
        self.height = height
        self.width = width
        self.eps = eps

    def hybrid_forward(self, F, points, K, T):
        P = F.batch_dot(K, T)[:, :3, :]

        cam_points = F.batch_dot(P, points)

        cam_pix = cam_points[:, :2, :] / (cam_points[:, 2, :].expand_dims(1) + self.eps)
        cam_pix = cam_pix.reshape(self.batch_size, 2, self.height, self.width)

        x_src = cam_pix[:, 0, :, :] / (self.width - 1)
        y_src = cam_pix[:, 1, :, :] / (self.height - 1)
        pix_coords = F.concat(x_src.expand_dims(1), y_src.expand_dims(1), dim=1)
        pix_coords = (pix_coords - 0.5) * 2

        return pix_coords

    def predict(self, points, K, T):
        P = mx.nd.batch_dot(K, T)[:, :3, :]

        cam_points = mx.nd.batch_dot(P, points)

        cam_pix = cam_points[:, :2, :] / (cam_points[:, 2, :].expand_dims(1) + self.eps)
        cam_pix = cam_pix.reshape(self.batch_size, 2, self.height, self.width)

        x_src = cam_pix[:, 0, :, :] / (self.width - 1)
        y_src = cam_pix[:, 1, :, :] / (self.height - 1)
        pix_coords = mx.nd.concat(x_src.expand_dims(1),
                                  y_src.expand_dims(1),
                                  dim=1)

        pix_coords = (pix_coords - 0.5) * 2

        return pix_coords


def get_smooth_loss(disp, img):
    """Computes the smoothness loss for a disparity image
    The color image is used for edge-aware smoothness
    """
    grad_disp_x = mx.nd.abs(disp[:, :, :, :-1] - disp[:, :, :, 1:])
    grad_disp_y = mx.nd.abs(disp[:, :, :-1, :] - disp[:, :, 1:, :])

    grad_img_x = mx.nd.mean(data=mx.nd.abs(img[:, :, :, :-1] - img[:, :, :, 1:]),
                            axis=1, keepdims=True)
    grad_img_y = mx.nd.mean(data=mx.nd.abs(img[:, :, :-1, :] - img[:, :, 1:, :]),
                            axis=1, keepdims=True)

    grad_disp_x = grad_disp_x * mx.nd.exp(-grad_img_x)
    grad_disp_y = grad_disp_y * mx.nd.exp(-grad_img_y)

    return grad_disp_x.mean() + grad_disp_y.mean()


class SSIM(nn.HybridBlock):
    """Layer to compute the SSIM loss between a pair of images
    """

    def __init__(self):
        super(SSIM, self).__init__()
        self.mu_x_pool = nn.AvgPool2D(3, 1)
        self.mu_y_pool = nn.AvgPool2D(3, 1)
        self.sig_x_pool = nn.AvgPool2D(3, 1)
        self.sig_y_pool = nn.AvgPool2D(3, 1)
        self.sig_xy_pool = nn.AvgPool2D(3, 1)

        self.refl = nn.ReflectionPad2D(1)

        self.C1 = 0.01 ** 2
        self.C2 = 0.03 ** 2

    def hybrid_forward(self, F, x, y):
        x = self.refl(x)
        y = self.refl(y)

        mu_x = self.mu_x_pool(x)
        mu_y = self.mu_y_pool(y)

        sigma_x = self.sig_x_pool(x ** 2) - mu_x ** 2
        sigma_y = self.sig_y_pool(y ** 2) - mu_y ** 2
        sigma_xy = self.sig_xy_pool(x * y) - mu_x * mu_y

        SSIM_n = (2 * mu_x * mu_y + self.C1) * (2 * sigma_xy + self.C2)
        SSIM_d = (mu_x ** 2 + mu_y ** 2 + self.C1) * (sigma_x + sigma_y + self.C2)

        return F.clip((1 - SSIM_n / SSIM_d) / 2, 0, 1)

    def predict(self, x, y):
        x = self.refl(x)
        y = self.refl(y)

        mu_x = self.mu_x_pool(x)
        mu_y = self.mu_y_pool(y)

        sigma_x = self.sig_x_pool(x ** 2) - mu_x ** 2
        sigma_y = self.sig_y_pool(y ** 2) - mu_y ** 2
        sigma_xy = self.sig_xy_pool(x * y) - mu_x * mu_y

        SSIM_n = (2 * mu_x * mu_y + self.C1) * (2 * sigma_xy + self.C2)
        SSIM_d = (mu_x ** 2 + mu_y ** 2 + self.C1) * (sigma_x + sigma_y + self.C2)

        return mx.nd.clip((1 - SSIM_n / SSIM_d) / 2, 0, 1)


def compute_depth_errors(gt, pred):
    """Computation of error metrics between predicted and ground truth depths
    """
    thresh = np.maximum((gt / pred), (pred / gt))
    a1 = (thresh < 1.25).mean()
    a2 = (thresh < 1.25 ** 2).mean()
    a3 = (thresh < 1.25 ** 3).mean()

    rmse = (gt - pred) ** 2
    rmse = np.sqrt(rmse.mean())

    rmse_log = (np.log(gt) - np.log(pred)) ** 2
    rmse_log = np.sqrt(rmse_log.mean())

    abs_rel = np.mean(np.abs(gt - pred) / gt)

    sq_rel = np.mean(((gt - pred) ** 2) / gt)

    return abs_rel, sq_rel, rmse, rmse_log, a1, a2, a3



# pylint: disable=wildcard-import
"""Video action recognition, something-something-v2 dataset.
https://20bn.com/datasets/something-something
"""
from __future__ import absolute_import
from .classification import *



"""Filesystem utility functions."""
import os
import errno
import contextlib
from pathlib import Path
import tarfile
import zipfile

def makedirs(path):
    """Create directory recursively if not exists.
    Similar to `makedir -p`, you can skip checking existence before this function.

    Parameters
    ----------
    path : str
        Path of the desired dir
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise

def try_import(package, message=None):
    """Try import specified package, with custom message support.

    Parameters
    ----------
    package : str
        The name of the targeting package.
    message : str, default is None
        If not None, this function will raise customized error message when import error is found.


    Returns
    -------
    module if found, raise ImportError otherwise

    """
    try:
        return __import__(package)
    except ImportError as e:
        if not message:
            raise e
        raise ImportError(message)

def try_import_cv2():
    """Try import cv2 at runtime.

    Returns
    -------
    cv2 module if found. Raise ImportError otherwise

    """
    msg = "cv2 is required, you can install by package manager, e.g. 'apt-get', \
        or `pip install opencv-python --user` (note that this is unofficial PYPI package)."
    return try_import('cv2', msg)

def try_import_colorama():
    """Try import colorama at runtime.

    Returns
    -------
    colorama module if found. Raise ImportError otherwise

    """
    msg = "colorama is required, you can install by `pip install colorama --user` \
         (note that this is unofficial PYPI package)."
    return try_import('colorama', msg)

def try_import_decord():
    """Try import decord at runtime.

    Returns
    -------
    Decord module if found. Raise ImportError otherwise

    """
    msg = "Decord is required, you can install by `pip install decord --user` \
        (note that this is unofficial PYPI package)."
    return try_import('decord', msg)

def try_import_mmcv():
    """Try import mmcv at runtime.

    Returns
    -------
    mmcv module if found. Raise ImportError otherwise

    """
    msg = "mmcv is required, you can install by first `pip install Cython --user` \
        and then `pip install mmcv --user` (note that this is unofficial PYPI package)."
    return try_import('mmcv', msg)

def try_import_rarfile():
    """Try import rarfile at runtime.

    Returns
    -------
    rarfile module if found. Raise ImportError otherwise

    """
    msg = "rarfile is required, you can install by first `sudo apt-get install unrar` \
        and then `pip install rarfile --user` (note that this is unofficial PYPI package)."
    return try_import('rarfile', msg)

def import_try_install(package, extern_url=None):
    """Try import the specified package.
    If the package not installed, try use pip to install and import if success.

    Parameters
    ----------
    package : str
        The name of the package trying to import.
    extern_url : str or None, optional
        The external url if package is not hosted on PyPI.
        For example, you can install a package using:
         "pip install git+http://github.com/user/repo/tarball/master/egginfo=xxx".
        In this case, you can pass the url to the extern_url.

    Returns
    -------
    <class 'Module'>
        The imported python module.

    """
    import tempfile
    import portalocker
    lockfile = os.path.join(tempfile.gettempdir(), package + '_install.lck')
    with portalocker.Lock(lockfile):
        try:
            return __import__(package)
        except ImportError:
            try:
                from pip import main as pipmain
            except ImportError:
                from pip._internal import main as pipmain
                from types import ModuleType
                # fix for pip 19.3
                if isinstance(pipmain, ModuleType):
                    from pip._internal.main import main as pipmain

            # trying to install package
            url = package if extern_url is None else extern_url
            pipmain(['install', '--user', url])  # will raise SystemExit Error if fails

            # trying to load again
            try:
                return __import__(package)
            except ImportError:
                import sys
                import site
                user_site = site.getusersitepackages()
                if user_site not in sys.path:
                    sys.path.append(user_site)
                return __import__(package)
    return __import__(package)

def try_import_dali():
    """Try import NVIDIA DALI at runtime.
    """
    try:
        dali = __import__('nvidia.dali', fromlist=['pipeline', 'ops', 'types'])
        dali.Pipeline = dali.pipeline.Pipeline
    except (ImportError, RuntimeError) as e:
        if isinstance(e, ImportError):
            msg = "DALI not found, please check if you installed it correctly."
        elif isinstance(e, RuntimeError):
            msg = "No CUDA-capable device is detected ({}).".format(e)
        class dali:
            class Pipeline:
                def __init__(self):
                    raise NotImplementedError(msg)
    return dali

def try_import_html5lib():
    """Try import html5lib at runtime.

    Returns
    -------
    html5lib module if found. Raise ImportError otherwise

    """
    msg = "html5lib is required, you can install by package manager, " \
          "e.g. pip install html5lib --user` (note that this is unofficial PYPI package)."
    return try_import('html5lib', msg)

def try_import_gdfDownloader():
    """Try import googleDriveFileDownloader at runtime.

    Returns
    -------
    googleDriveFileDownloader module if found. Raise ImportError otherwise

    """
    msg = "googleDriveFileDownloader is required, you can install by package manager, " \
          "e.g. pip install googleDriveFileDownloader --user` " \
          "(note that this is unofficial PYPI package)."
    return try_import('googleDriveFileDownloader', msg)

def unzip(zip_file_path, root='./', strict=False):
    """Unzips files located at `zip_file_path` into parent directory specified by `root`.
    """
    root = os.path.expanduser(root)
    with zipfile.ZipFile(zip_file_path) as zf:
        if strict or not os.path.exists(os.path.join(root, zf.namelist()[-1])):
            zf.extractall(root)
        folder = os.path.commonprefix(zf.namelist())
    return os.path.join(root, folder)

def untar(tar_file_path, root='./', strict=False):
    """Untars files located at `tar_file_path` into parent directory specified by `root`.
    """
    root = os.path.expanduser(root)
    with tarfile.open(tar_file_path, 'r:gz') as zf:
        if strict or not os.path.exists(os.path.join(root, zf.getnames()[-1])):
            zf.extractall(root)
        folder = os.path.commonprefix(zf.getnames())
    return os.path.join(root, folder)

@contextlib.contextmanager
def temporary_filename(suffix=None):
    """Context that introduces a temporary file.

    Creates a temporary file, yields its name, and upon context exit, deletes it.
    (In contrast, tempfile.NamedTemporaryFile() provides a 'file' object and
    deletes the file as soon as that file object is closed, so the temporary file
    cannot be safely re-opened by another library or process.)

    Parameters
    ----------
    suffix: desired filename extension (e.g. '.mp4').

    Yields
    ----------
    The name of the temporary file.
    """
    import tempfile
    try:
        f = tempfile.NamedTemporaryFile(suffix=suffix, delete=False)
        tmp_name = f.name
        f.close()
        yield tmp_name
    finally:
        os.unlink(tmp_name)

class _DisplayablePath:
    """A util class for displaying the tree structure of root path.

    Example:

    >>> paths = _DisplayablePath.make_tree(Path('doc'))
    >>> for path in paths:
    >>>    print(path.displayable())

    Parameters
    ----------
    path : str
        The path.
    parent_path : str
        The parent parth.
    is_last : bool
        Whether it's the last node in this depth.

    """
    display_filename_prefix_middle = '├──'
    display_filename_prefix_last = '└──'
    display_parent_prefix_middle = '    '
    display_parent_prefix_last = '│   '

    def __init__(self, path, parent_path, is_last):
        self.path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    # pylint: disable=inconsistent-return-statements
    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None, max_depth=1):
        """Make tree structure from root.

        Parameters
        ----------
        root : str
            The root dir.
        parent : _DisplayablePath
            The parent displayable path.
        is_last : bool
            Whether it's the last in this level.
        criteria : function
            The criteria used to filter dir/path.
        max_depth : int
            Maximum depth for search.

        """
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        if displayable_root.depth > max_depth:
            return displayable_root
        yield displayable_root

        children = sorted(list(path
                               for path in root.iterdir()
                               if criteria(path)),
                          key=lambda s: str(s).lower())
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir() and displayable_root.depth < max_depth - 1:
                yield from cls.make_tree(path,
                                         parent=displayable_root,
                                         is_last=is_last,
                                         criteria=criteria,
                                         max_depth=max_depth)
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        _ = path
        return True

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + '/'
        return self.path.name

    def displayable(self):
        """Display string"""
        if self.parent is None:
            return self.displayname

        _filename_prefix = (self.display_filename_prefix_last
                            if self.is_last
                            else self.display_filename_prefix_middle)

        parts = ['{!s} {!s}'.format(_filename_prefix,
                                    self.displayname)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(self.display_parent_prefix_middle
                         if parent.is_last
                         else self.display_parent_prefix_last)
            parent = parent.parent

        return ''.join(reversed(parts))


class PathTree:
    """A directory tree structure viewer.

    Parameters
    ----------
    root : str or pathlib.Path
        The root directory.
    max_depth : int
        Max depth for recursive sub-folders, please be conservative to not spam the filesystem.

    """
    def __init__(self, root, max_depth=1):
        self._disp_path = _DisplayablePath.make_tree(Path(root), max_depth=max_depth)

    def __str__(self):
        s = '\n'.join([p.displayable() for p in self._disp_path])
        return s

def try_import_skimage():
    """Try import scikit-image at runtime.

    Returns
    -------
    scikit-image module if found. Raise ImportError otherwise

    """
    msg = "skimage is required, you can install by package manager, e.g. " \
          "`pip install scikit-image --user` (note that this is unofficial PYPI package)."
    return try_import('skimage', msg)



from __future__ import print_function
from __future__ import division

import unittest
import mxnet as mx
import numpy as np

import gluoncv as gcv

def test_viz_bbox():
    img = mx.nd.zeros((300, 300, 3), dtype=np.uint8)
    bbox = mx.nd.array([[10, 20, 200, 500], [150, 200, 400, 300]])
    scores = mx.nd.array([0.8, 0.001])
    labels = mx.nd.array([1, 3])
    class_names = ['a', 'b', 'c']
    ax = gcv.utils.viz.plot_bbox(img, bbox, scores=scores, labels=labels, class_names=class_names)
    ax = gcv.utils.viz.plot_bbox(img, bbox, ax=ax, reverse_rgb=True)
    ax = gcv.utils.viz.plot_bbox(img, bbox / 500, ax=ax, reverse_rgb=True, absolute_coordinates=False)

    img_output = gcv.utils.viz.cv_plot_bbox(img, bbox, scores=scores, labels=labels, class_names=class_names)
    img_output = gcv.utils.viz.cv_plot_bbox(img, bbox)
    img_output = gcv.utils.viz.cv_plot_bbox(img, bbox / 500, absolute_coordinates=False)

def test_viz_image():
    img = mx.nd.zeros((300, 300, 3), dtype=np.uint8)
    ax = gcv.utils.viz.plot_image(img)
    ax = gcv.utils.viz.plot_image(img, ax=ax, reverse_rgb=True)

    # No X Server
    # img = gcv.utils.viz.cv_plot_image(img)

def test_viz_keypoints():
    img = mx.nd.zeros((300, 300, 3), dtype=np.uint8)
    bbox = mx.nd.array([[[10, 20, 200, 500], [150, 200, 400, 300]]])
    scores = mx.nd.array([[0.8, 0.001]])
    labels = mx.nd.array([[1, 3]])
    class_names = ['a', 'b', 'c']

    coords = np.array([[[100, 100],
                        [100, 101],
                        [100, 102],
                        [100, 103],
                        [100, 104],
                        [100, 105],
                        [100, 106],
                        [100, 107],
                        [100, 108],
                        [100, 109],
                        [100, 110],
                        [100, 111],
                        [100, 112],
                        [100, 113],
                        [100, 114],
                        [100, 115],
                        [100, 116]]]).astype('float32')
    confidence = np.array([[[0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9],
                            [0.9]]])

    class_IDs = mx.nd.array([[[0], [0]]])
    scores = mx.nd.array([[[1], [1]]])
    img_output = gcv.utils.viz.plot_keypoints(img, coords, confidence, class_IDs, bbox, scores)
    img_output = gcv.utils.viz.cv_plot_keypoints(img, coords, confidence, class_IDs, bbox, scores)

@unittest.skip("Skip due to graphviz env")
def test_viz_network():
    try:
        import graphviz
        for name in ['mobilenet1.0', 'resnet50_v1b']:
            net = gcv.model_zoo.get_model(name, pretrained=True)
            for shape in [(1, 3, 224, 224), (1, 3, 448, 448)]:
                gcv.utils.viz.plot_network(net)
    except ImportError:
        pass

if __name__ == '__main__':
    import nose
    nose.runmodule()



""" tracking Data augmentation

Code adapted from https://github.com/STVIR/pysot
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from gluoncv.utils.filesystem import try_import_cv2
from gluoncv.model_zoo.siamrpn.siamrpn_tracker import corner2center, center2corner
from gluoncv.model_zoo.siamrpn.siamrpn_tracker import Center, Corner
class SiamRPNaugmentation:
    """dataset Augmentation for SiamRPN tracking.

    Parameters
    ----------
    shift : int
        length of template augmentation shift
    scale : float
        template augmentation scale ratio
    blur : float
        template augmentation blur ratio
    flip : float
        template augmentation flip ratio
    color : float
        template augmentation color ratio
    """
    def __init__(self, shift, scale, blur, flip, color):
        self.shift = shift
        self.scale = scale
        self.blur = blur
        self.flip = flip
        self.color = color
        self.rgbVar = np.array([[-0.55919361, 0.98062831, - 0.41940627],
                                [1.72091413, 0.19879334, - 1.82968581],
                                [4.64467907, 4.73710203, 4.88324118]], dtype=np.float32)
        self.cv2 = try_import_cv2()

    @staticmethod
    def random():
        return np.random.random() * 2 - 1.0

    def _crop_roi(self, image, bbox, out_sz, padding=(0, 0, 0)):
        """crop image roi size.

        Parameters
        ----------
        image : np.array
            image
        bbox : list or np.array
            bbox coordinate，like (xmin,ymin,xmax,ymax)
        out_sz : int
            size after crop

        Return:
            image after crop
        """
        bbox = [float(x) for x in bbox]
        a = (out_sz-1) / (bbox[2]-bbox[0])
        b = (out_sz-1) / (bbox[3]-bbox[1])
        c = -a * bbox[0]
        d = -b * bbox[1]
        mapping = np.array([[a, 0, c],
                            [0, b, d]]).astype(np.float)
        crop = self.cv2.warpAffine(image, mapping, (out_sz, out_sz),
                                   borderMode=self.cv2.BORDER_CONSTANT,
                                   borderValue=padding)
        return crop

    def _blur_aug(self, image):
        """blur filter to smooth image

        Parameters
        ----------
        image : np.array
            image

        Return:
            image after blur
        """
        def rand_kernel():
            sizes = np.arange(5, 46, 2)
            size = np.random.choice(sizes)
            kernel = np.zeros((size, size))
            c = int(size/2)
            wx = np.random.random()
            kernel[:, c] += 1. / size * wx
            kernel[c, :] += 1. / size * (1-wx)
            return kernel
        kernel = rand_kernel()
        image = self.cv2.filter2D(image, -1, kernel)
        return image

    def _color_aug(self, image):
        """Random increase of image channel

        Parameters
        ----------
        image : np.array
            image

        Return:
            image after Random increase of channel
        """
        offset = np.dot(self.rgbVar, np.random.randn(3, 1))
        offset = offset[::-1]  # bgr 2 rgb
        offset = offset.reshape(3)
        image = image - offset
        return image

    def _gray_aug(self, image):
        """image Grayscale

        Parameters
        ----------
        image : np.array
            image

        Return:
            image after Grayscale
        """
        grayed = self.cv2.cvtColor(image, self.cv2.COLOR_BGR2GRAY)
        image = self.cv2.cvtColor(grayed, self.cv2.COLOR_GRAY2BGR)
        return image

    def _shift_scale_aug(self, image, bbox, crop_bbox, size):
        """shift scale augmentation

        Parameters
        ----------
        image : np.array
            image
        bbox : list or np.array
            bbox
        crop_bbox :
            crop size image from center

        Return
            image ,bbox after shift and scale
        """
        im_h, im_w = image.shape[:2]

        # adjust crop bounding box
        crop_bbox_center = corner2center(crop_bbox)

        if self.scale:
            scale_x = (1.0 + SiamRPNaugmentation.random() * self.scale)
            scale_y = (1.0 + SiamRPNaugmentation.random() * self.scale)
            h, w = crop_bbox_center.h, crop_bbox_center.w
            scale_x = min(scale_x, float(im_w) / w)
            scale_y = min(scale_y, float(im_h) / h)
            crop_bbox_center = Center(crop_bbox_center.x,
                                      crop_bbox_center.y,
                                      crop_bbox_center.w * scale_x,
                                      crop_bbox_center.h * scale_y)

        crop_bbox = center2corner(crop_bbox_center)
        if self.shift:
            sx = SiamRPNaugmentation.random() * self.shift
            sy = SiamRPNaugmentation.random() * self.shift

            x1, y1, x2, y2 = crop_bbox

            sx = max(-x1, min(im_w - 1 - x2, sx))
            sy = max(-y1, min(im_h - 1 - y2, sy))

            crop_bbox = Corner(x1 + sx, y1 + sy, x2 + sx, y2 + sy)

        # adjust target bounding box
        x1, y1 = crop_bbox.x1, crop_bbox.y1
        bbox = Corner(bbox.x1 - x1, bbox.y1 - y1,
                      bbox.x2 - x1, bbox.y2 - y1)

        if self.scale:
            bbox = Corner(bbox.x1 / scale_x, bbox.y1 / scale_y,
                          bbox.x2 / scale_x, bbox.y2 / scale_y)

        image = self._crop_roi(image, crop_bbox, size)
        return image, bbox

    def _flip_aug(self, image, bbox):
        """flip augmentation

        Parameters
        ----------
        image : np.array
            image
        bbox : list or np.array
            bbox

        Return
            image and bbox after filp
        """
        image = self.cv2.flip(image, 1)
        width = image.shape[1]
        bbox = Corner(width - 1 - bbox.x2, bbox.y1,
                      width - 1 - bbox.x1, bbox.y2)
        return image, bbox

    def __call__(self, image, bbox, size, gray=False):
        shape = image.shape
        crop_bbox = center2corner(Center(shape[0]//2, shape[1]//2,
                                         size-1, size-1))
        # gray augmentation
        if gray:
            image = self._gray_aug(image)

        # shift scale augmentation
        image, bbox = self._shift_scale_aug(image, bbox, crop_bbox, size)

        # color augmentation
        if self.color > np.random.random():
            image = self._color_aug(image)

        # blur augmentation
        if self.blur > np.random.random():
            image = self._blur_aug(image)

        # flip augmentation
        if self.flip and self.flip > np.random.random():
            image, bbox = self._flip_aug(image, bbox)
        return image, bbox



# coding: utf-8

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import print_function

import warnings
import logging

import mxnet as mx
from mxnet.contrib.quantization import *
from ..unittests.common import try_gpu, with_cpu

import gluoncv as gcv


def test_get_all_models():
    names = gcv.model_zoo.get_model_list()
    for name in names:
        kwargs = {}
        if 'custom' in name:
            kwargs['classes'] = ['a', 'b']
        if 'dcnv2' in name and not hasattr(mx.gluon.contrib.cnn, 'ModulatedDeformableConvolution'):
            continue
        net = gcv.model_zoo.get_model(name, pretrained=False, **kwargs)
        assert isinstance(net, mx.gluon.Block), '{}'.format(name)


def _test_model_list(model_list, ctx, x, pretrained=True, **kwargs):
    pretrained_models = gcv.model_zoo.pretrained_model_list()
    for model in model_list:
        if model in pretrained_models:
            net = gcv.model_zoo.get_model(model, pretrained=True, **kwargs)
        else:
            net = gcv.model_zoo.get_model(model, **kwargs)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                net.initialize()
        net.collect_params().reset_ctx(ctx)
        net(x)
        mx.nd.waitall()

def _calib_model_list(model_list, ctx, x, pretrained=True, **kwargs):
    pretrained_models = gcv.model_zoo.pretrained_model_list()
    for model in model_list:
        if model in pretrained_models:
            net = gcv.model_zoo.get_model(model, pretrained=True, **kwargs)
        else:
            net = gcv.model_zoo.get_model(model, **kwargs)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                net.initialize()
        net.collect_params().reset_ctx(ctx)
        exclude_layers_match = ['flatten']
        if model.find('ssd') != -1 or model.find('psp') != -1 or model.find('deeplab') != -1:
            exclude_layers_match += ['concat']
        random_label = mx.random.uniform(shape=(x.shape[0],1))
        dataset = mx.gluon.data.dataset.ArrayDataset(x, random_label)
        calib_data = mx.gluon.data.DataLoader(dataset, batch_size=1)
        dummy_logger = logging.getLogger('dummy')
        dummy_logger.setLevel(logging.ERROR)
        net = quantize_net(net, quantized_dtype='auto',
                           exclude_layers=None,
                           exclude_layers_match=exclude_layers_match,
                           calib_data=calib_data, calib_mode='naive',
                           num_calib_examples=1, ctx=ctx, logger=dummy_logger)
        net(x)
        mx.nd.waitall()

@with_cpu(0)
def _test_bn_global_stats(model_list, **kwargs):
    class _BatchNorm(mx.gluon.nn.BatchNorm):
        def __init__(self, axis=1, momentum=0.9, epsilon=1e-5, center=True, scale=True,
                     use_global_stats=False, beta_initializer='zeros', gamma_initializer='ones',
                     running_mean_initializer='zeros', running_variance_initializer='ones',
                     in_channels=0, **kwargs):
            assert use_global_stats
            super(_BatchNorm, self).__init__(axis, momentum, epsilon, center, scale,
                                             use_global_stats, beta_initializer, gamma_initializer,
                                             running_mean_initializer, running_variance_initializer,
                                             in_channels, **kwargs)

    for model in model_list:
        gcv.model_zoo.get_model(model, norm_layer=_BatchNorm, use_global_stats=True, **kwargs)


@try_gpu(0)
def test_classification_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(2, 3, 32, 32), ctx=ctx)
    cifar_models = [
        'cifar_resnet20_v1', 'cifar_resnet56_v1', 'cifar_resnet110_v1',
        'cifar_resnet20_v2', 'cifar_resnet56_v2', 'cifar_resnet110_v2',
        'cifar_wideresnet16_10', 'cifar_wideresnet28_10', 'cifar_wideresnet40_8',
        'cifar_resnext29_32x4d', 'cifar_resnext29_16x64d',
        'cifar_residualattentionnet56', 'cifar_residualattentionnet92',
        'cifar_residualattentionnet452'
    ]
    _test_model_list(cifar_models, ctx, x)


@try_gpu(0)
def test_imagenet_models():
    ctx = mx.context.current_context()

    # 224x224
    x = mx.random.uniform(shape=(2, 3, 224, 224), ctx=ctx)
    models = ['resnet18_v1', 'resnet34_v1',
              'resnet50_v1', 'resnet101_v1', 'resnet152_v1',
              'resnet18_v1b', 'resnet34_v1b', 'resnet50_v1b_gn',
              'resnet50_v1b', 'resnet101_v1b', 'resnet152_v1b',
              'resnet50_v1c', 'resnet101_v1c', 'resnet152_v1c',
              'resnet50_v1d', 'resnet101_v1d', 'resnet152_v1d',
              'resnet50_v1e', 'resnet101_v1e', 'resnet152_v1e',
              'resnet18_v2', 'resnet34_v2',
              'resnet50_v2', 'resnet101_v2', 'resnet152_v2',
              'resnest50', 'resnest101', 'resnest200', 'resnest269',
              'resnext50_32x4d', 'resnext101_32x4d', 'resnext101_64x4d',
              'se_resnext50_32x4d', 'se_resnext101_32x4d', 'se_resnext101_64x4d',
              'se_resnet18_v1', 'se_resnet34_v1', 'se_resnet50_v1',
              'se_resnet101_v1', 'se_resnet152_v1',
              'se_resnet18_v2', 'se_resnet34_v2', 'se_resnet50_v2',
              'se_resnet101_v2', 'se_resnet152_v2',
              'senet_154', 'squeezenet1.0', 'squeezenet1.1',
              'mobilenet1.0', 'mobilenet0.75', 'mobilenet0.5', 'mobilenet0.25',
              'mobilenetv2_1.0', 'mobilenetv2_0.75', 'mobilenetv2_0.5', 'mobilenetv2_0.25',
              'mobilenetv3_large', 'mobilenetv3_small',
              'densenet121', 'densenet161', 'densenet169', 'densenet201',
              'darknet53', 'alexnet', 'googlenet',
              'vgg11', 'vgg11_bn', 'vgg13', 'vgg13_bn',
              'vgg16', 'vgg16_bn', 'vgg19', 'vgg19_bn',
              'residualattentionnet56', 'residualattentionnet92',
              'residualattentionnet128', 'residualattentionnet164',
              'residualattentionnet200', 'residualattentionnet236', 'residualattentionnet452',
              'resnet18_v1b_0.89', 'resnet50_v1d_0.86', 'resnet50_v1d_0.48',
              'resnet50_v1d_0.37', 'resnet50_v1d_0.11',
              'resnet101_v1d_0.76', 'resnet101_v1d_0.73']
    _test_model_list(models, ctx, x)

    # 299x299
    x = mx.random.uniform(shape=(2, 3, 299, 299), ctx=ctx)
    models = ['inceptionv3', 'nasnet_5_1538', 'nasnet_7_1920', 'nasnet_6_4032', 'xception']
    _test_model_list(models, ctx, x)

    # 331x331
    x = mx.random.uniform(shape=(2, 3, 331, 331), ctx=ctx)
    models = ['nasnet_5_1538', 'nasnet_7_1920', 'nasnet_6_4032']
    _test_model_list(models, ctx, x)


@try_gpu(0)
def test_simple_pose_resnet_models():
    ctx = mx.context.current_context()
    models = ['simple_pose_resnet18_v1b',
              'simple_pose_resnet50_v1b', 'simple_pose_resnet101_v1b', 'simple_pose_resnet152_v1b',
              'simple_pose_resnet50_v1d', 'simple_pose_resnet101_v1d', 'simple_pose_resnet152_v1d',
              'mobile_pose_resnet18_v1b', 'mobile_pose_resnet50_v1b',
              'mobile_pose_mobilenet1.0', 'mobile_pose_mobilenetv2_1.0',
              'mobile_pose_mobilenetv3_large', 'mobile_pose_mobilenetv3_small']

    # 192x256
    x = mx.random.uniform(shape=(2, 3, 192, 256), ctx=ctx)
    _test_model_list(models, ctx, x)

    # 256x256
    x = mx.random.uniform(shape=(2, 3, 256, 256), ctx=ctx)
    _test_model_list(models, ctx, x)

    # 288x384
    x = mx.random.uniform(shape=(2, 3, 288, 384), ctx=ctx)
    _test_model_list(models, ctx, x)

@try_gpu(0)
def test_alpha_pose_resnet_models():
    ctx = mx.context.current_context()
    models = ['alpha_pose_resnet101_v1b_coco']

    # 256x320
    x = mx.random.uniform(shape=(2, 3, 256, 320), ctx=ctx)
    _test_model_list(models, ctx, x)


def test_imagenet_models_bn_global_stats():
    models = ['resnet18_v1b', 'resnet34_v1b', 'resnet50_v1b',
              'resnet101_v1b', 'resnet152_v1b']
    _test_bn_global_stats(models)


def test_ssd_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)  # allow non-squre and larger inputs
    models = ['ssd_300_vgg16_atrous_voc', 'ssd_512_vgg16_atrous_voc', 'ssd_512_resnet50_v1_voc']
    if not mx.context.current_context().device_type == 'gpu':
        models = ['ssd_512_resnet50_v1_voc']
    _test_model_list(models, ctx, x)

def test_center_net_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 512), ctx=ctx)  # allow non-squre and larger inputs
    models = ['center_net_resnet18_v1b_voc', 'center_net_resnet50_v1b_coco']
    if not mx.context.current_context().device_type == 'gpu':
        models = ['center_net_resnet18_v1b_voc']
    if hasattr(mx.gluon.contrib.cnn, 'ModulatedDeformableConvolution'):
        models.append('center_net_resnet18_v1b_dcnv2_voc')
    _test_model_list(models, ctx, x)

def test_ssd_reset_class():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)  # allow non-squre and larger inputs
    model_name = 'ssd_300_vgg16_atrous_voc'
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"], reuse_weights=["bus", "car", "bird"])
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"], reuse_weights={"bus": "bus"})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={"person": 14})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={0: 14})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={0: "person"})
    test_classes = ['bird', 'bicycle', 'bus', 'car', 'cat']
    test_classes_dict = dict(zip(test_classes, test_classes))
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(test_classes, reuse_weights=test_classes_dict)
    net(x)


# This test is only executed when a gpu is available
def test_ssd_reset_class_on_gpu():
    ctx = mx.gpu(0)
    try:
        x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)
    except Exception:
        return

    model_name = 'ssd_300_vgg16_atrous_voc'
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"], reuse_weights=["bus", "car", "bird"])
    net(x)


def test_yolo3_reset_class():
    test_classes = ['bird', 'bicycle', 'bus', 'car', 'cat']
    test_classes_dict = dict(zip(test_classes, test_classes))

    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)  # allow non-square and larger inputs
    model_name = 'yolo3_darknet53_voc'
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.hybridize()
    net.reset_class(["bus", "car", "bird"], reuse_weights=["bus", "car", "bird"])
    net(x)
    mx.nd.waitall()

    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.hybridize()
    net.reset_class(test_classes, reuse_weights=test_classes_dict)
    net(x)
    mx.nd.waitall()

    # for GPU
    ctx = mx.gpu(0)
    try:
        x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)
    except Exception:
        return
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.hybridize()
    net.reset_class(["bus", "car", "bird"])
    net(x)
    mx.nd.waitall()

    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.hybridize()
    net.reset_class(test_classes, reuse_weights=test_classes_dict)
    net(x)
    mx.nd.waitall()


def test_faster_rcnn_reset_class():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)  # allow non-squre and larger inputs
    model_name = 'faster_rcnn_resnet50_v1b_coco'
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"], reuse_weights=["bus", "car", "bird"])
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"], reuse_weights={"bus": "bus"})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={"person": 14})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={0: 14})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={0: "person"})
    net(x)

    # for GPU
    ctx = mx.gpu(0)
    try:
        x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)
    except Exception:
        return
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"])
    net(x)


def test_mask_rcnn_reset_class():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)  # allow non-squre and larger inputs
    model_name = 'mask_rcnn_resnet50_v1b_coco'
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"], reuse_weights=["bus", "car", "bird"])
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"], reuse_weights={"bus": "bus"})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={"person": 14})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={0: 14})
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["person", "car", "bird"], reuse_weights={0: "person"})
    net(x)

    # for GPU
    ctx = mx.gpu(0)
    try:
        x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)
    except Exception:
        return
    net = gcv.model_zoo.get_model(model_name, pretrained=True, ctx=ctx)
    net.reset_class(["bus", "car", "bird"])
    net(x)


@try_gpu(0)
def test_faster_rcnn_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 300, 400), ctx=ctx)  # allow non-squre and larger inputs
    models = ['faster_rcnn_resnet50_v1b_voc', 'faster_rcnn_resnet50_v1b_coco',
              'faster_rcnn_fpn_resnet50_v1b_coco']
    _test_model_list(models, ctx, x)


@try_gpu(0)
def test_mask_rcnn_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 300, 400), ctx=ctx)
    models = ['mask_rcnn_resnet50_v1b_coco', 'mask_rcnn_fpn_resnet50_v1b_coco',
              'mask_rcnn_resnet18_v1b_coco', 'mask_rcnn_fpn_resnet18_v1b_coco']
    _test_model_list(models, ctx, x)


@try_gpu(0)
def test_rcnn_max_dets_greater_than_nms_mask_rcnn_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 300, 400), ctx=ctx)
    net = gcv.model_zoo.mask_rcnn_resnet18_v1b_coco(pretrained=False, pretrained_base=True,
                                                    rcnn_max_dets=1000, rpn_test_pre_nms=100,
                                                    rpn_test_post_nms=30)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        net.initialize()
    net.collect_params().reset_ctx(ctx)
    net(x)
    mx.nd.waitall()


def test_yolo3_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 320, 320), ctx=ctx)  # allow non-squre and larger inputs
    models = ['yolo3_darknet53_voc']
    _test_model_list(models, ctx, x)


@try_gpu(0)
def test_two_stage_ctx_loading():
    model_name = 'yolo3_darknet53_coco'
    ctx = mx.test_utils.default_context()
    if str(ctx).startswith('cpu'):
        # meaningless
        return
    net = gcv.model_zoo.get_model(model_name, pretrained=True)
    net.save_parameters(model_name + '.params')
    net = gcv.model_zoo.get_model(model_name, pretrained=False, ctx=ctx)
    net.load_parameters(model_name + '.params', ctx=ctx)


def test_set_nms():
    model_list = ['ssd_512_resnet50_v1_voc', 'faster_rcnn_resnet50_v1b_voc', 'yolo3_darknet53_coco']
    for model in model_list:
        net = gcv.model_zoo.get_model(model, pretrained=False, pretrained_base=False)
        net.initialize()
        net.hybridize()
        ctx = mx.context.current_context()
        x = mx.random.uniform(shape=(1, 3, 608, 768), ctx=ctx)
        net.set_nms(nms_thresh=0.45, nms_topk=400, post_nms=100)
        net(x)
        net.set_nms(nms_thresh=0.3, nms_topk=200, post_nms=50)
        net(x)


@try_gpu(0)
def test_segmentation_models():
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 480, 480), ctx=ctx)
    models = ['fcn_resnet101_coco', 'psp_resnet101_coco', 'deeplab_resnet101_coco',
              'fcn_resnet101_voc', 'psp_resnet101_voc', 'deeplab_resnet101_voc',
              'fcn_resnet50_ade', 'psp_resnet50_ade', 'deeplab_resnet50_ade',
              'fcn_resnet101_ade', 'psp_resnet101_ade', 'deeplab_resnet101_ade',
              'deeplab_resnest200_ade',
              'psp_resnet101_citys', 'deeplab_resnet152_voc', 'deeplab_resnet152_coco',
              'deeplab_v3b_plus_wideresnet_citys', 'icnet_resnet50_citys',
              'deeplab_resnet50_citys', 'deeplab_resnet101_citys']
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)


@try_gpu(0)
def test_segmentation_models_custom_size():
    ctx = mx.context.current_context()
    num_classes = 5
    width = 96
    height = 64
    x = mx.random.uniform(shape=(1, 3, height, width), ctx=ctx)

    net = gcv.model_zoo.FCN(num_classes, backbone='resnet50', aux=False, ctx=ctx, pretrained_base=True,
                            height=height, width=width)
    result = net.forward(x)
    assert result[0].shape == (1, num_classes, height, width)
    net = gcv.model_zoo.PSPNet(num_classes, backbone='resnet50', aux=False, ctx=ctx, pretrained_base=True,
                               height=height, width=width)
    result = net.forward(x)
    assert result[0].shape == (1, num_classes, height, width)

    net = gcv.model_zoo.DeepLabV3(num_classes, backbone='resnet50', aux=False, ctx=ctx, pretrained_base=True,
                               height=height, width=width)
    result = net.forward(x)
    assert result[0].shape == (1, num_classes, height, width)

    net = gcv.model_zoo.DeepLabV3Plus(num_classes, backbone='resnet50', aux=False, ctx=ctx, pretrained_base=True,
                                  height=height, width=width)
    result = net.forward(x)
    assert result[0].shape == (1, num_classes, height, width)

    net = gcv.model_zoo.DeepLabWV3Plus(num_classes, backbone='resnet50', aux=False, ctx=ctx, pretrained_base=True,
                                  height=height, width=width)
    result = net.forward(x)
    assert result[0].shape == (1, num_classes, height, width)

@try_gpu(0)
def test_action_recognition_vgg_models():
    ctx = mx.context.current_context()
    models = ['vgg16_ucf101']

    # 224x224
    x = mx.random.uniform(shape=(2, 3, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_inceptionv1_models():
    ctx = mx.context.current_context()
    models = ['inceptionv1_kinetics400']

    # 299x299
    x = mx.random.uniform(shape=(2, 3, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_inceptionv3_models():
    ctx = mx.context.current_context()
    models = ['inceptionv3_ucf101', 'inceptionv3_kinetics400']

    # 299x299
    x = mx.random.uniform(shape=(2, 3, 299, 299), ctx=ctx)    # only allow 299x299 input
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_resnet_models():
    ctx = mx.context.current_context()
    models = ['resnet18_v1b_kinetics400', 'resnet34_v1b_kinetics400', 'resnet50_v1b_kinetics400',
              'resnet101_v1b_kinetics400', 'resnet152_v1b_kinetics400', 'resnet50_v1b_sthsthv2',
              'resnet50_v1b_hmdb51']

    # 224x224
    x = mx.random.uniform(shape=(2, 3, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_c3d_models():
    ctx = mx.context.current_context()
    models = ['c3d_kinetics400']

    # 112x112
    x = mx.random.uniform(shape=(2, 3, 16, 112, 112), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_p3d_models():
    ctx = mx.context.current_context()
    models = ['p3d_resnet50_kinetics400', 'p3d_resnet101_kinetics400']

    # 224x224
    x = mx.random.uniform(shape=(2, 3, 16, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_r2plus1d_models():
    ctx = mx.context.current_context()
    models = ['r2plus1d_resnet18_kinetics400', 'r2plus1d_resnet34_kinetics400',
              'r2plus1d_resnet50_kinetics400']

    # 112x112
    x = mx.random.uniform(shape=(2, 3, 16, 112, 112), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_i3d_models():
    ctx = mx.context.current_context()
    models = ['i3d_resnet50_v1_kinetics400', 'i3d_resnet101_v1_kinetics400', 'i3d_inceptionv1_kinetics400',
              'i3d_inceptionv3_kinetics400', 'i3d_resnet50_v1_sthsthv2', 'i3d_resnet50_v1_hmdb51',
              'i3d_resnet50_v1_ucf101', 'i3d_nl5_resnet50_v1_kinetics400', 'i3d_nl10_resnet50_v1_kinetics400',
              'i3d_nl5_resnet101_v1_kinetics400', 'i3d_nl10_resnet101_v1_kinetics400']

    # 224x224
    x = mx.random.uniform(shape=(2, 3, 32, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@try_gpu(0)
def test_action_recognition_slowfast_models():
    ctx = mx.context.current_context()

    models = ['slowfast_4x16_resnet50_kinetics400']
    x = mx.random.uniform(shape=(2, 3, 36, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

    models = ['slowfast_8x8_resnet50_kinetics400']
    x = mx.random.uniform(shape=(2, 3, 40, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

    models = ['slowfast_8x8_resnet101_kinetics400']
    x = mx.random.uniform(shape=(2, 3, 40, 224, 224), ctx=ctx)
    _test_model_list(models, ctx, x, pretrained=True, pretrained_base=True)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=False)
    _test_model_list(models, ctx, x, pretrained=False, pretrained_base=True)

@with_cpu(0)
def test_mobilenet_sync_bn():
    model_name = "mobilenet1.0"
    net = gcv.model_zoo.get_model(model_name, pretrained=True)
    net.save_parameters(model_name + '.params')
    net = gcv.model_zoo.get_model(model_name, pretrained=False,
                                  norm_layer=mx.gluon.contrib.nn.SyncBatchNorm,
                                  norm_kwargs={'num_devices': 2})
    net.load_parameters(model_name + '.params')

@with_cpu(0)
def test_quantized_imagenet_models():
    model_list = ['mobilenet1.0_int8', 'resnet50_v1_int8']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 224, 224), ctx=ctx)
    _test_model_list(model_list, ctx, x)

@with_cpu(0)
def test_quantized_ssd_models():
    model_list = ['ssd_300_vgg16_atrous_voc_int8', 'ssd_512_mobilenet1.0_voc_int8',
                  'ssd_512_resnet50_v1_voc_int8', 'ssd_512_vgg16_atrous_voc_int8']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)
    _test_model_list(model_list, ctx, x)

@with_cpu(0)
def test_calib_models():
    model_list = ['resnet50_v1', 'resnet50_v1d_0.11',
                  'mobilenet1.0', 'mobilenetv2_1.0',
                  'squeezenet1.0', 'squeezenet1.1',
                  'vgg16']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 224, 224), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

    model_list = ['inceptionv3']
    x = mx.random.uniform(shape=(1, 3, 299, 299), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

    model_list = ['ssd_300_vgg16_atrous_voc', 'ssd_512_mobilenet1.0_voc',
                  'ssd_512_resnet50_v1_voc', 'ssd_512_vgg16_atrous_voc']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 512, 544), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

    model_list = ['fcn_resnet101_voc', 'fcn_resnet101_coco',
                  'psp_resnet101_voc', 'psp_resnet101_coco',
                  'deeplab_resnet101_voc', 'deeplab_resnet101_coco']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 480, 480), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

    model_list = ['simple_pose_resnet18_v1b',
                  'simple_pose_resnet50_v1b', 'simple_pose_resnet50_v1d',
                  'simple_pose_resnet101_v1b', 'simple_pose_resnet101_v1d']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 256, 192), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

    model_list = ['vgg16_ucf101', 'resnet18_v1b_kinetics400', 'resnet50_v1b_kinetics400']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 224, 224), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

    model_list = ['inceptionv3_ucf101', 'inceptionv3_kinetics400']
    x = mx.random.uniform(shape=(1, 3, 299, 299), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

    model_list = ['yolo3_darknet53_voc', 'yolo3_darknet53_coco',
                  'yolo3_mobilenet1.0_voc', 'yolo3_mobilenet1.0_coco']
    x = mx.random.uniform(shape=(1, 3, 416, 416), ctx=ctx)
    _calib_model_list(model_list, ctx, x)

@with_cpu(0)
def test_quantized_segmentation_models():
    model_list = ['fcn_resnet101_voc_int8', 'fcn_resnet101_coco_int8',
                  'psp_resnet101_voc_int8', 'psp_resnet101_coco_int8',
                  'deeplab_resnet101_voc_int8', 'deeplab_resnet101_coco_int8']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 480, 480), ctx=ctx)
    _test_model_list(model_list, ctx, x)


@with_cpu(0)
def test_quantized_pose_estimation_models():
    model_list = ['simple_pose_resnet18_v1b_int8',
                  'simple_pose_resnet50_v1b_int8', 'simple_pose_resnet50_v1d_int8',
                  'simple_pose_resnet101_v1b_int8', 'simple_pose_resnet101_v1d_int8']
    ctx = mx.context.current_context()
    x = mx.random.uniform(shape=(1, 3, 256, 192), ctx=ctx)
    _test_model_list(model_list, ctx, x)


if __name__ == '__main__':
    import nose

    nose.runmodule()



# pylint: disable=missing-docstring, unused-argument, line-too-long, too-many-lines
"""
Temporal Pyramid Network for Action Recognition
CVPR 2020, https://arxiv.org/pdf/2004.03548.pdf
Code adapted from https://github.com/decisionforce/TPN
"""
import numpy as np

import torch
import torch.nn as nn
import torch.utils.checkpoint as cp
import torch.nn.functional as F


__all__ = ['TPN', 'TPNet', 'tpn_resnet50_f8s8_kinetics400', 'tpn_resnet50_f16s4_kinetics400',
           'tpn_resnet50_f32s2_kinetics400', 'tpn_resnet101_f8s8_kinetics400',
           'tpn_resnet101_f16s4_kinetics400', 'tpn_resnet101_f32s2_kinetics400',
           'tpn_resnet50_f32s2_custom']


def conv3x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "3x3x3 convolution with padding"
    return nn.Conv3d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=(temporal_stride, spatial_stride, spatial_stride),
        padding=dilation,
        dilation=dilation,
        bias=False)


def conv1x3x3(in_planes, out_planes, spatial_stride=1, temporal_stride=1, dilation=1):
    "1x3x3 convolution with padding"
    return nn.Conv3d(
        in_planes,
        out_planes,
        kernel_size=(1, 3, 3),
        stride=(temporal_stride, spatial_stride, spatial_stride),
        padding=(0, dilation, dilation),
        dilation=dilation,
        bias=False)


class Identity(nn.Module):

    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


class ConvModule(nn.Module):
    def __init__(
            self,
            inplanes,
            planes,
            kernel_size,
            stride,
            padding,
            bias=False,
            groups=1,
    ):
        super(ConvModule, self).__init__()
        self.conv = nn.Conv3d(inplanes, planes, kernel_size, stride, padding, bias=bias, groups=groups)
        self.bn = nn.BatchNorm3d(planes)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        out = self.relu(self.bn(self.conv(x)))
        return out


class AuxHead(nn.Module):
    def __init__(self, inplanes, planes, loss_weight=0.5):
        super(AuxHead, self).__init__()
        self.convs = ConvModule(inplanes, inplanes * 2, kernel_size=(1, 3, 3),
                                stride=(1, 2, 2), padding=(0, 1, 1), bias=False)
        self.loss_weight = loss_weight
        self.dropout = nn.Dropout(p=0.5)
        self.fc = nn.Linear(inplanes * 2, planes)

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)
            if isinstance(m, nn.Conv3d):
                pass
            if isinstance(m, nn.BatchNorm3d):
                m.weight.data.fill_(1)
                m.bias.data.fill_(0)

    def forward(self, x, target=None):
        if target is None:
            return None
        loss = dict()
        x = self.convs(x)
        x = F.adaptive_avg_pool3d(x, 1).squeeze(-1).squeeze(-1).squeeze(-1)
        x = self.dropout(x)
        x = self.fc(x)

        loss['loss_aux'] = self.loss_weight * F.cross_entropy(x, target)
        return loss


class TemporalModulation(nn.Module):
    def __init__(self, inplanes, planes, downsample_scale=8):
        super(TemporalModulation, self).__init__()

        self.conv = nn.Conv3d(inplanes, planes, (3, 1, 1), (1, 1, 1), (1, 0, 0), bias=False, groups=32)
        self.pool = nn.MaxPool3d((downsample_scale, 1, 1), (downsample_scale, 1, 1), (0, 0, 0), ceil_mode=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.pool(x)
        return x


class Upsampling(nn.Module):
    def __init__(self, scale=(2, 1, 1)):
        super(Upsampling, self).__init__()
        self.scale = scale

    def forward(self, x):
        x = F.interpolate(x, scale_factor=self.scale, mode='nearest')
        return x


class Downampling(nn.Module):
    def __init__(self,
                 inplanes,
                 planes,
                 kernel_size=(3, 1, 1),
                 stride=(1, 1, 1),
                 padding=(1, 0, 0),
                 bias=False,
                 groups=1,
                 norm=False,
                 activation=False,
                 downsample_position='after',
                 downsample_scale=(1, 2, 2),
                 ):
        super(Downampling, self).__init__()

        self.conv = nn.Conv3d(inplanes, planes, kernel_size, stride, padding, bias=bias, groups=groups)
        self.norm = nn.BatchNorm3d(planes) if norm else None
        self.relu = nn.ReLU(inplace=True) if activation else None
        assert (downsample_position in ['before', 'after'])
        self.downsample_position = downsample_position
        self.pool = nn.MaxPool3d(downsample_scale, downsample_scale, (0, 0, 0), ceil_mode=True)

    def forward(self, x):
        # pylint: disable=not-callable
        if self.downsample_position == 'before':
            x = self.pool(x)
        x = self.conv(x)
        if self.norm is not None:
            x = self.norm(x)
        if self.relu is not None:
            x = self.relu(x)
        if self.downsample_position == 'after':
            x = self.pool(x)

        return x


class LevelFusion(nn.Module):
    def __init__(self,
                 in_channels=(1024, 1024),
                 mid_channels=(1024, 1024),
                 out_channels=2048,
                 ds_scales=((1, 1, 1), (1, 1, 1)),
                 ):
        super(LevelFusion, self).__init__()
        self.ops = nn.ModuleList()
        num_ins = len(in_channels)
        for i in range(num_ins):
            op = Downampling(in_channels[i], mid_channels[i], kernel_size=(1, 1, 1), stride=(1, 1, 1),
                             padding=(0, 0, 0), bias=False, groups=32, norm=True, activation=True,
                             downsample_position='before', downsample_scale=ds_scales[i])
            self.ops.append(op)

        in_dims = np.sum(mid_channels)
        self.fusion_conv = nn.Sequential(
            nn.Conv3d(in_dims, out_channels, 1, 1, 0, bias=False),
            nn.BatchNorm3d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, inputs):
        out = [self.ops[i](feature) for i, feature in enumerate(inputs)]
        out = torch.cat(out, 1)
        out = self.fusion_conv(out)
        return out


class SpatialModulation(nn.Module):
    def __init__(self, inplanes=(1024, 2048), planes=2048):
        super(SpatialModulation, self).__init__()

        self.spatial_modulation = nn.ModuleList()
        for _, dim in enumerate(inplanes):
            op = nn.ModuleList()
            ds_factor = planes // dim
            ds_num = int(np.log2(ds_factor))
            if ds_num < 1:
                op = Identity()
            else:
                for dsi in range(ds_num):
                    in_factor = 2 ** dsi
                    out_factor = 2 ** (dsi + 1)
                    op.append(ConvModule(dim * in_factor, dim * out_factor, kernel_size=(1, 3, 3), stride=(1, 2, 2),
                                         padding=(0, 1, 1), bias=False))
            self.spatial_modulation.append(op)

    def forward(self, inputs):
        out = []
        for i, _ in enumerate(inputs):
            if isinstance(self.spatial_modulation[i], nn.ModuleList):
                out_ = inputs[i]
                for _, op in enumerate(self.spatial_modulation[i]):
                    out_ = op(out_)
                out.append(out_)
            else:
                out.append(self.spatial_modulation[i](inputs[i]))
        return out


class TPN(nn.Module):

    def __init__(self,
                 in_channels=(256, 512, 1024, 2048),
                 out_channels=256,
                 spatial_modulation_config=None,
                 temporal_modulation_config=None,
                 upsampling_config=None,
                 downsampling_config=None,
                 level_fusion_config=None,
                 aux_head_config=None,
                 ):
        super(TPN, self).__init__()
        assert isinstance(out_channels, int)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_ins = len(in_channels)

        self.temporal_modulation_ops = nn.ModuleList()
        self.upsampling_ops = nn.ModuleList()
        self.downsampling_ops = nn.ModuleList()
        self.level_fusion_op = LevelFusion(**level_fusion_config)
        self.spatial_modulation = SpatialModulation(**spatial_modulation_config)
        for i in range(0, self.num_ins, 1):
            inplanes = in_channels[-1]
            planes = out_channels

            if temporal_modulation_config is not None:
                # overwrite the temporal_modulation_config
                temporal_modulation_config['param']['downsample_scale'] = temporal_modulation_config['scales'][i]
                temporal_modulation_config['param']['inplanes'] = inplanes
                temporal_modulation_config['param']['planes'] = planes
                temporal_modulation = TemporalModulation(**temporal_modulation_config['param'])
                self.temporal_modulation_ops.append(temporal_modulation)

            if i < self.num_ins - 1:
                if upsampling_config is not None:
                    # overwrite the upsampling_config
                    upsampling = Upsampling(**upsampling_config)
                    self.upsampling_ops.append(upsampling)

                if downsampling_config is not None:
                    # overwrite the downsampling_config
                    downsampling_config['param']['inplanes'] = planes
                    downsampling_config['param']['planes'] = planes
                    downsampling_config['param']['downsample_scale'] = downsampling_config['scales']
                    downsampling = Downampling(**downsampling_config['param'])
                    self.downsampling_ops.append(downsampling)

        out_dims = level_fusion_config['out_channels']

        # Two pyramids
        self.level_fusion_op2 = LevelFusion(**level_fusion_config)

        self.pyramid_fusion_op = nn.Sequential(
            nn.Conv3d(out_dims * 2, 2048, 1, 1, 0, bias=False),
            nn.BatchNorm3d(2048),
            nn.ReLU(inplace=True)
        )

        # overwrite aux_head_config
        if aux_head_config is not None:
            aux_head_config['inplanes'] = self.in_channels[-2]
            self.aux_head = AuxHead(**aux_head_config)
        else:
            self.aux_head = None

    # default init_weights for conv(msra) and norm in ConvModule
    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                pass
            if isinstance(m, nn.BatchNorm3d):
                m.weight.data.fill_(1)
                m.bias.data.fill_(0)

        if self.aux_head is not None:
            self.aux_head.init_weights()

    def forward(self, inputs, target=None):
        loss = None

        # Auxiliary loss
        if self.aux_head is not None:
            loss = self.aux_head(inputs[-2], target)

        # Spatial Modulation
        outs = self.spatial_modulation(inputs)

        # Temporal Modulation
        outs = [temporal_modulation(outs[i]) for i, temporal_modulation in enumerate(self.temporal_modulation_ops)]

        temporal_modulation_outs = outs

        # Build top-down flow - upsampling operation
        if self.upsampling_ops is not None:
            for i in range(self.num_ins - 1, 0, -1):
                outs[i - 1] = outs[i - 1] + self.upsampling_ops[i - 1](outs[i])

        # Get top-down outs
        topdownouts = self.level_fusion_op2(outs)
        outs = temporal_modulation_outs

        # Build bottom-up flow - downsampling operation
        if self.downsampling_ops is not None:
            for i in range(0, self.num_ins - 1, 1):
                outs[i + 1] = outs[i + 1] + self.downsampling_ops[i](outs[i])

        # Get bottom-up outs
        outs = self.level_fusion_op(outs)

        # fuse two pyramid outs
        outs = self.pyramid_fusion_op(torch.cat([topdownouts, outs], 1))

        return outs, loss


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 spatial_stride=1,
                 temporal_stride=1,
                 dilation=1,
                 downsample=None,
                 style='pytorch',
                 if_inflate=True,
                 inflate_style='3x1x1',
                 if_nonlocal=True,
                 nonlocal_cfg=None,
                 with_cp=False):
        """Bottleneck block for ResNet.
        If style is "pytorch", the stride-two layer is the 3x3 conv layer,
        if it is "caffe", the stride-two layer is the first 1x1 conv layer.
        """
        super(Bottleneck, self).__init__()
        assert style in ['pytorch', 'caffe']
        assert inflate_style in ['3x1x1', '3x3x3']
        self.inplanes = inplanes
        self.planes = planes

        if style == 'pytorch':
            self.conv1_stride = 1
            self.conv2_stride = spatial_stride
            self.conv1_stride_t = 1
            self.conv2_stride_t = temporal_stride
        else:
            self.conv1_stride = spatial_stride
            self.conv2_stride = 1
            self.conv1_stride_t = temporal_stride
            self.conv2_stride_t = 1
        if if_inflate:
            if inflate_style == '3x1x1':
                self.conv1 = nn.Conv3d(
                    inplanes,
                    planes,
                    kernel_size=(3, 1, 1),
                    stride=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                    padding=(1, 0, 0),
                    bias=False)
                self.conv2 = nn.Conv3d(
                    planes,
                    planes,
                    kernel_size=(1, 3, 3),
                    stride=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                    padding=(0, dilation, dilation),
                    dilation=(1, dilation, dilation),
                    bias=False)
            else:
                self.conv1 = nn.Conv3d(
                    inplanes,
                    planes,
                    kernel_size=1,
                    stride=(self.conv1_stride_t, self.conv1_stride, self.conv1_stride),
                    bias=False)
                self.conv2 = nn.Conv3d(
                    planes,
                    planes,
                    kernel_size=3,
                    stride=(self.conv2_stride_t, self.conv2_stride, self.conv2_stride),
                    padding=(1, dilation, dilation),
                    dilation=(1, dilation, dilation),
                    bias=False)
        else:
            self.conv1 = nn.Conv3d(
                inplanes,
                planes,
                kernel_size=1,
                stride=(1, self.conv1_stride, self.conv1_stride),
                bias=False)
            self.conv2 = nn.Conv3d(
                planes,
                planes,
                kernel_size=(1, 3, 3),
                stride=(1, self.conv2_stride, self.conv2_stride),
                padding=(0, dilation, dilation),
                dilation=(1, dilation, dilation),
                bias=False)

        self.bn1 = nn.BatchNorm3d(planes)
        self.bn2 = nn.BatchNorm3d(planes)
        self.conv3 = nn.Conv3d(
            planes, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm3d(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.spatial_tride = spatial_stride
        self.temporal_tride = temporal_stride
        self.dilation = dilation
        self.with_cp = with_cp

        if if_nonlocal and nonlocal_cfg is not None:
            nonlocal_cfg_ = nonlocal_cfg.copy()
            nonlocal_cfg_['in_channels'] = planes * self.expansion
            self.nonlocal_block = build_nonlocal_block(nonlocal_cfg_)
        else:
            self.nonlocal_block = None

    def forward(self, x):

        def _inner_forward(x):
            identity = x

            out = self.conv1(x)
            out = self.bn1(out)
            out = self.relu(out)

            out = self.conv2(out)
            out = self.bn2(out)
            out = self.relu(out)

            out = self.conv3(out)
            out = self.bn3(out)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.relu(out)

        if self.nonlocal_block is not None:
            out = self.nonlocal_block(out)

        return out


def make_res_layer(block,
                   inplanes,
                   planes,
                   blocks,
                   spatial_stride=1,
                   temporal_stride=1,
                   dilation=1,
                   style='pytorch',
                   inflate_freq=1,
                   inflate_style='3x1x1',
                   nonlocal_freq=1,
                   nonlocal_cfg=None,
                   with_cp=False):
    inflate_freq = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * blocks
    nonlocal_freq = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * blocks
    assert len(inflate_freq) == blocks
    assert len(nonlocal_freq) == blocks
    downsample = None
    if spatial_stride != 1 or inplanes != planes * block.expansion:
        downsample = nn.Sequential(
            nn.Conv3d(
                inplanes,
                planes * block.expansion,
                kernel_size=1,
                stride=(temporal_stride, spatial_stride, spatial_stride),
                bias=False),
            nn.BatchNorm3d(planes * block.expansion),
        )

    layers = []
    layers.append(
        block(
            inplanes,
            planes,
            spatial_stride,
            temporal_stride,
            dilation,
            downsample,
            style=style,
            if_inflate=(inflate_freq[0] == 1),
            inflate_style=inflate_style,
            if_nonlocal=(nonlocal_freq[0] == 1),
            nonlocal_cfg=nonlocal_cfg,
            with_cp=with_cp))
    inplanes = planes * block.expansion
    for i in range(1, blocks):
        layers.append(
            block(inplanes,
                  planes,
                  1, 1,
                  dilation,
                  style=style,
                  if_inflate=(inflate_freq[i] == 1),
                  inflate_style=inflate_style,
                  if_nonlocal=(nonlocal_freq[i] == 1),
                  nonlocal_cfg=nonlocal_cfg,
                  with_cp=with_cp))

    return nn.Sequential(*layers)


class TPNet(nn.Module):
    """ResNe(x)t_SlowFast backbone.
    Args:
        depth (int): Depth of resnet, from {50, 101, 152}.
        num_stages (int): Resnet stages, normally 4.
        strides (Sequence[int]): Strides of the first block of each stage.
        dilations (Sequence[int]): Dilation of each stage.
        out_indices (Sequence[int]): Output from which stages.
        style (str): `pytorch` or `caffe`. If set to "pytorch", the stride-two
            layer is the 3x3 conv layer, otherwise the stride-two layer is
            the first 1x1 conv layer.
        frozen_stages (int): Stages to be frozen (all param fixed). -1 means
            not freezing any parameters.
        bn_eval (bool): Whether to set BN layers to eval mode, namely, freeze
            running stats (mean and var).
        bn_frozen (bool): Whether to freeze weight and bias of BN layers.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed.
    """

    arch_settings = {
        50: (Bottleneck, (3, 4, 6, 3)),
        101: (Bottleneck, (3, 4, 23, 3)),
        152: (Bottleneck, (3, 8, 36, 3))
    }

    def __init__(self,
                 depth,
                 TPN_neck=None,
                 num_classes=400,
                 pretrained=None,
                 pretrained_base=True,
                 feat_ext=False,
                 num_stages=4,
                 spatial_strides=(1, 2, 2, 2),
                 temporal_strides=(1, 1, 1, 1),
                 dilations=(1, 1, 1, 1),
                 out_indices=(2, 3),
                 conv1_kernel_t=1,
                 conv1_stride_t=1,
                 pool1_kernel_t=1,
                 pool1_stride_t=1,
                 style='pytorch',
                 frozen_stages=-1,
                 inflate_freq=(0, 0, 1, 1),
                 inflate_stride=(1, 1, 1, 1),
                 inflate_style='3x1x1',
                 nonlocal_stages=(-1,),
                 nonlocal_freq=(0, 0, 0, 0),
                 nonlocal_cfg=None,
                 bn_eval=False,
                 bn_frozen=False,
                 partial_bn=False,
                 with_cp=False,
                 dropout_ratio=0.5):
        super(TPNet, self).__init__()
        if depth not in self.arch_settings:
            raise KeyError('invalid depth {} for resnet'.format(depth))
        self.depth = depth
        self.pretrained = pretrained
        self.pretrained_base = pretrained_base
        self.num_classes = num_classes
        self.num_stages = num_stages
        assert 1 <= num_stages <= 4
        self.spatial_strides = spatial_strides
        self.temporal_strides = temporal_strides
        self.dilations = dilations
        assert len(spatial_strides) == len(temporal_strides) == len(dilations) == num_stages
        self.out_indices = out_indices
        assert max(out_indices) < num_stages
        self.style = style
        self.frozen_stages = frozen_stages
        self.inflate_freqs = inflate_freq if not isinstance(inflate_freq, int) else (inflate_freq,) * num_stages
        self.inflate_style = inflate_style
        self.nonlocal_stages = nonlocal_stages
        self.nonlocal_freqs = nonlocal_freq if not isinstance(nonlocal_freq, int) else (nonlocal_freq,) * num_stages
        self.nonlocal_cfg = nonlocal_cfg
        self.bn_eval = bn_eval
        self.bn_frozen = bn_frozen
        self.partial_bn = partial_bn
        self.with_cp = with_cp
        self.feat_ext = feat_ext

        self.dropout_ratio = dropout_ratio

        self.block, stage_blocks = self.arch_settings[depth]
        self.stage_blocks = stage_blocks[:num_stages]
        self.inplanes = 64

        self.conv1 = nn.Conv3d(
            3, 64, kernel_size=(conv1_kernel_t, 7, 7), stride=(conv1_stride_t, 2, 2),
            padding=((conv1_kernel_t - 1) // 2, 3, 3), bias=False)
        self.bn1 = nn.BatchNorm3d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=(pool1_kernel_t, 3, 3), stride=(pool1_stride_t, 2, 2),
                                    padding=(pool1_kernel_t // 2, 1, 1))

        self.res_layers = []
        for i, num_blocks in enumerate(self.stage_blocks):
            spatial_stride = spatial_strides[i]
            temporal_stride = temporal_strides[i]
            dilation = dilations[i]
            planes = 64 * 2 ** i
            res_layer = make_res_layer(
                self.block,
                self.inplanes,
                planes,
                num_blocks,
                spatial_stride=spatial_stride,
                temporal_stride=temporal_stride,
                dilation=dilation,
                style=self.style,
                inflate_freq=self.inflate_freqs[i],
                inflate_style=self.inflate_style,
                nonlocal_freq=self.nonlocal_freqs[i],
                nonlocal_cfg=self.nonlocal_cfg if i in self.nonlocal_stages else None,
                with_cp=with_cp)
            self.inplanes = planes * self.block.expansion
            layer_name = 'layer{}'.format(i + 1)
            self.add_module(layer_name, res_layer)
            self.res_layers.append(layer_name)

        self.feat_dim = self.block.expansion * 64 * 2 ** (len(self.stage_blocks) - 1)

        self.TPN_neck = TPN_neck

        if self.dropout_ratio != 0:
            self.dropout = nn.Dropout(p=self.dropout_ratio)
        else:
            self.dropout = None
        self.avg_pool = nn.AdaptiveAvgPool3d(1)
        self.fc = nn.Linear(2048, self.num_classes)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x1 = self.layer1(x)
        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x4 = self.layer4(x3)

        feat = [x3, x4]
        tpn_feat = self.TPN_neck(feat)

        x = self.avg_pool(tpn_feat[0])
        if self.dropout is not None:
            x = self.dropout(x)
        x = x.view(x.size(0), -1)

        if self.feat_ext:
            return x

        out = self.fc(x)
        return out


def tpn_resnet50_f8s8_kinetics400(cfg):
    neck = TPN(in_channels=[1024, 2048],
               out_channels=1024,
               spatial_modulation_config=dict(
                   inplanes=[1024, 2048],
                   planes=2048, ),
               temporal_modulation_config=dict(
                   scales=(8, 8),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               upsampling_config=dict(
                   scale=(1, 1, 1),
               ),
               downsampling_config=dict(
                   scales=(1, 1, 1),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               level_fusion_config=dict(
                   in_channels=[1024, 1024],
                   mid_channels=[1024, 1024],
                   out_channels=2048,
                   ds_scales=[(1, 1, 1), (1, 1, 1)],
               ),
               aux_head_config=dict(
                   inplanes=-1,
                   planes=cfg.CONFIG.DATA.NUM_CLASSES,
                   loss_weight=0.5
               ))

    model = TPNet(depth=50,
                  TPN_neck=neck,
                  num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                  pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                  pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                  feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                  bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                  partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                  bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('tpn_resnet50_f8s8_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def tpn_resnet50_f16s4_kinetics400(cfg):
    neck = TPN(in_channels=[1024, 2048],
               out_channels=1024,
               spatial_modulation_config=dict(
                   inplanes=[1024, 2048],
                   planes=2048, ),
               temporal_modulation_config=dict(
                   scales=(16, 16),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               upsampling_config=dict(
                   scale=(1, 1, 1),
               ),
               downsampling_config=dict(
                   scales=(1, 1, 1),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               level_fusion_config=dict(
                   in_channels=[1024, 1024],
                   mid_channels=[1024, 1024],
                   out_channels=2048,
                   ds_scales=[(1, 1, 1), (1, 1, 1)],
               ),
               aux_head_config=dict(
                   inplanes=-1,
                   planes=cfg.CONFIG.DATA.NUM_CLASSES,
                   loss_weight=0.5
               ))

    model = TPNet(depth=50,
                  TPN_neck=neck,
                  num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                  pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                  pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                  feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                  bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                  partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                  bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('tpn_resnet50_f16s4_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def tpn_resnet50_f32s2_kinetics400(cfg):
    neck = TPN(in_channels=[1024, 2048],
               out_channels=1024,
               spatial_modulation_config=dict(
                   inplanes=[1024, 2048],
                   planes=2048, ),
               temporal_modulation_config=dict(
                   scales=(32, 32),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               upsampling_config=dict(
                   scale=(1, 1, 1),
               ),
               downsampling_config=dict(
                   scales=(1, 1, 1),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               level_fusion_config=dict(
                   in_channels=[1024, 1024],
                   mid_channels=[1024, 1024],
                   out_channels=2048,
                   ds_scales=[(1, 1, 1), (1, 1, 1)],
               ),
               aux_head_config=dict(
                   inplanes=-1,
                   planes=cfg.CONFIG.DATA.NUM_CLASSES,
                   loss_weight=0.5
               ))

    model = TPNet(depth=50,
                  TPN_neck=neck,
                  num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                  pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                  pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                  feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                  bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                  partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                  bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('tpn_resnet50_f32s2_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def tpn_resnet101_f8s8_kinetics400(cfg):
    neck = TPN(in_channels=[1024, 2048],
               out_channels=1024,
               spatial_modulation_config=dict(
                   inplanes=[1024, 2048],
                   planes=2048, ),
               temporal_modulation_config=dict(
                   scales=(4, 8),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               upsampling_config=dict(
                   scale=(1, 1, 1),
               ),
               downsampling_config=dict(
                   scales=(2, 1, 1),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               level_fusion_config=dict(
                   in_channels=[1024, 1024],
                   mid_channels=[1024, 1024],
                   out_channels=2048,
                   ds_scales=[(2, 1, 1), (1, 1, 1)],
               ),
               aux_head_config=dict(
                   inplanes=-1,
                   planes=cfg.CONFIG.DATA.NUM_CLASSES,
                   loss_weight=0.5
               ))

    model = TPNet(depth=101,
                  TPN_neck=neck,
                  num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                  pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                  pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                  feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                  bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                  partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                  bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('tpn_resnet101_f8s8_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def tpn_resnet101_f16s4_kinetics400(cfg):
    neck = TPN(in_channels=[1024, 2048],
               out_channels=1024,
               spatial_modulation_config=dict(
                   inplanes=[1024, 2048],
                   planes=2048, ),
               temporal_modulation_config=dict(
                   scales=(8, 16),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               upsampling_config=dict(
                   scale=(1, 1, 1),
               ),
               downsampling_config=dict(
                   scales=(2, 1, 1),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               level_fusion_config=dict(
                   in_channels=[1024, 1024],
                   mid_channels=[1024, 1024],
                   out_channels=2048,
                   ds_scales=[(2, 1, 1), (1, 1, 1)],
               ),
               aux_head_config=dict(
                   inplanes=-1,
                   planes=cfg.CONFIG.DATA.NUM_CLASSES,
                   loss_weight=0.5
               ))

    model = TPNet(depth=101,
                  TPN_neck=neck,
                  num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                  pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                  pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                  feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                  bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                  partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                  bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('tpn_resnet101_f16s4_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def tpn_resnet101_f32s2_kinetics400(cfg):
    neck = TPN(in_channels=[1024, 2048],
               out_channels=1024,
               spatial_modulation_config=dict(
                   inplanes=[1024, 2048],
                   planes=2048, ),
               temporal_modulation_config=dict(
                   scales=(16, 32),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               upsampling_config=dict(
                   scale=(1, 1, 1),
               ),
               downsampling_config=dict(
                   scales=(2, 1, 1),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               level_fusion_config=dict(
                   in_channels=[1024, 1024],
                   mid_channels=[1024, 1024],
                   out_channels=2048,
                   ds_scales=[(2, 1, 1), (1, 1, 1)],
               ),
               aux_head_config=dict(
                   inplanes=-1,
                   planes=cfg.CONFIG.DATA.NUM_CLASSES,
                   loss_weight=0.5
               ))

    model = TPNet(depth=101,
                  TPN_neck=neck,
                  num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                  pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                  pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                  feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                  bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                  partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                  bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('tpn_resnet101_f32s2_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def tpn_resnet50_f32s2_custom(cfg):
    neck = TPN(in_channels=[1024, 2048],
               out_channels=1024,
               spatial_modulation_config=dict(
                   inplanes=[1024, 2048],
                   planes=2048, ),
               temporal_modulation_config=dict(
                   scales=(32, 32),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               upsampling_config=dict(
                   scale=(1, 1, 1),
               ),
               downsampling_config=dict(
                   scales=(1, 1, 1),
                   param=dict(
                       inplanes=-1,
                       planes=-1,
                       downsample_scale=-1,
                   )),
               level_fusion_config=dict(
                   in_channels=[1024, 1024],
                   mid_channels=[1024, 1024],
                   out_channels=2048,
                   ds_scales=[(1, 1, 1), (1, 1, 1)],
               ),
               aux_head_config=dict(
                   inplanes=-1,
                   planes=cfg.CONFIG.DATA.NUM_CLASSES,
                   loss_weight=0.5
               ))

    model = TPNet(depth=50,
                  TPN_neck=neck,
                  num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                  pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                  pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                  feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                  bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                  partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                  bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        state_dict = torch.load(get_model_file('tpn_resnet50_f32s2_kinetics400', tag=cfg.CONFIG.MODEL.PRETRAINED))
        for k in list(state_dict.keys()):
            # retain only backbone up to before the classification layer
            if k.startswith('fc'):
                del state_dict[k]

        msg = model.load_state_dict(state_dict, strict=False)
        assert set(msg.missing_keys) == {'fc.weight', 'fc.bias'}
        print("=> initialized from a SlowFast4x16 model pretrained on Kinetcis400 dataset")
    return model



"""Encoder module of Monodepth2
Code partially borrowed from
https://github.com/nianticlabs/monodepth2/blob/master/networks/resnet_encoder.py
"""
from __future__ import absolute_import, division, print_function

import os
import numpy as np
import mxnet as mx

from mxnet.gluon import nn
from mxnet.context import cpu
from ...model_zoo.resnetv1b import \
    resnet18_v1b, resnet34_v1b, resnet50_v1s, resnet101_v1s, resnet152_v1s


class ResnetEncoder(nn.HybridBlock):
    r"""Encoder of Monodepth2

    Parameters
    ----------
    backbone : string
        Pre-trained dilated backbone network type ('resnet18', 'resnet34', 'resnet50',
        'resnet101' or 'resnet152').
    pretrained : bool or str
        Refers to if the backbone is pretrained or not. If `True`,
        model weights of a model that was trained on ImageNet is loaded.
    num_input_images : int
        The number of input sequences. 1 for depth encoder, larger than 1 for pose encoder.
        (Default: 1)
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    """
    def __init__(self, backbone, pretrained, num_input_images=1,
                 root=os.path.join(os.path.expanduser('~'), '.mxnet/models'),
                 ctx=cpu(), **kwargs):
        super(ResnetEncoder, self).__init__()

        self.num_ch_enc = np.array([64, 64, 128, 256, 512])

        resnets = {'resnet18': resnet18_v1b,
                   'resnet34': resnet34_v1b,
                   'resnet50': resnet50_v1s,
                   'resnet101': resnet101_v1s,
                   'resnet152': resnet152_v1s}

        num_layers = {'resnet18': 18,
                      'resnet34': 34,
                      'resnet50': 50,
                      'resnet101': 101,
                      'resnet152': 152}

        if backbone not in resnets:
            raise ValueError("{} is not a valid resnet".format(backbone))

        if num_input_images > 1:
            self.encoder = resnets[backbone](pretrained=False, ctx=ctx, **kwargs)
            if pretrained:
                filename = os.path.join(
                    root, 'resnet%d_v%db_multiple_inputs.params' % (num_layers[backbone], 1))
                if not os.path.isfile(filename):
                    from ..model_store import get_model_file
                    loaded = mx.nd.load(get_model_file('resnet%d_v%db' % (num_layers[backbone], 1),
                                                       tag=pretrained, root=root))
                    loaded['conv1.weight'] = mx.nd.concat(
                        *([loaded['conv1.weight']] * num_input_images), dim=1) / num_input_images
                    mx.nd.save(filename, loaded)
                self.encoder.load_parameters(filename, ctx=ctx)
                from ...data import ImageNet1kAttr
                attrib = ImageNet1kAttr()
                self.encoder.synset = attrib.synset
                self.encoder.classes = attrib.classes
                self.encoder.classes_long = attrib.classes_long
        else:
            self.encoder = resnets[backbone](pretrained=pretrained, ctx=ctx, **kwargs)

        if backbone not in ('resnet18', 'resnet34'):
            self.num_ch_enc[1:] *= 4

    def hybrid_forward(self, F, input_image):
        # pylint: disable=unused-argument, missing-function-docstring
        self.features = []
        x = (input_image - 0.45) / 0.225
        x = self.encoder.conv1(x)
        x = self.encoder.bn1(x)
        self.features.append(self.encoder.relu(x))
        self.features.append(self.encoder.layer1(self.encoder.maxpool(self.features[-1])))
        self.features.append(self.encoder.layer2(self.features[-1]))
        self.features.append(self.encoder.layer3(self.features[-1]))
        self.features.append(self.encoder.layer4(self.features[-1]))

        return self.features

    def predict(self, input_image):
        # pylint: disable=unused-argument, missing-function-docstring
        self.features = []
        x = (input_image - 0.45) / 0.225
        x = self.encoder.conv1(x)
        x = self.encoder.bn1(x)
        self.features.append(self.encoder.relu(x))
        self.features.append(self.encoder.layer1(self.encoder.maxpool(self.features[-1])))
        self.features.append(self.encoder.layer2(self.features[-1]))
        self.features.append(self.encoder.layer3(self.features[-1]))
        self.features.append(self.encoder.layer4(self.features[-1]))

        return self.features
